from lionfuncs import to_dict

from lion_core.communication.action_request import ActionRequest
from lion_core.generic.note import note
from lion_core.session.msg_handlers.extract_request import (
    extract_action_request,
    extract_request_from_content_code_block,
    extract_request_plain_function_calling,
)


def create_action_request(response: str | dict) -> list[ActionRequest] | None:
    msg = note(**to_dict(response))

    content_ = None
    if str(msg.get(["content"], "")).strip().lower() == "none":
        content_ = extract_request_plain_function_calling(msg, suppress=True)
    elif a := msg.get(["content", "tool_uses"], None):
        content_ = a
    else:
        content_ = extract_request_from_content_code_block(msg)

    if content_:
        content_ = [content_] if isinstance(content_, dict) else content_
        return extract_action_request(content_)

    _content = to_dict(msg["content"], fuzzy_parse=True, suppress=True)

    if _content and isinstance(_content, dict):
        if "action_request" in _content:
            content_ = _content["action_request"]
        if isinstance(content_, str):
            content_ = to_dict(content_, fuzzy_parse=True, suppress=True)
        if isinstance(content_, dict):
            content_ = [content_]
        if isinstance(content_, list):
            return extract_action_request(content_)

    return None



from typing import Any

from lionabc import Signal
from pydantic import Field
from typing_extensions import override

from lion_core.communication.mail import Mail
from lion_core.communication.package import Package
from lion_core.generic.element import Element
from lion_core.generic.exchange import Exchange


class StartMail(Element, Signal):
    """A start mail node that triggers the initiation of a process."""

    mailbox: Exchange = Field(
        default_factory=Exchange, description="The pending start mail"
    )

    @override
    def trigger(
        self,
        context: Any,
        structure_id: str,
        executable_id: str,
    ) -> None:
        """Triggers the start mail by including it in the mailbox.

        Args:
            context: The context to be included in the start mail.
            structure_id: The ID of the structure to be initiated.
            executable_id: The ID of the executable to receive the start mail.
        """
        start_mail_content = {"context": context, "structure_id": structure_id}
        pack = Package(category="start", package=start_mail_content)
        start_mail = Mail(
            sender=self.ln_id,
            recipient=executable_id,
            package=pack,
        )
        self.mailbox.include(start_mail, "out")


# File: lion_core/communication/start_mail.py



from collections.abc import Sequence
from typing import Any, Protocol, TypeVar, runtime_checkable

from lionabc.exceptions import LionTypeError, LionValueError

T = TypeVar("T")


@runtime_checkable
class PileLoader(Protocol[T]):
    """Protocol defining the interface for pile loader classes."""

    @staticmethod
    def from_obj(cls, obj: T) -> dict | Sequence[dict]: ...

    @staticmethod
    def can_load(cls, obj: Any) -> bool: ...


class PileLoaderRegistry:
    _loaders: dict[str, PileLoader] = {}

    @classmethod
    def register(cls, key: str, loader: PileLoader | type[PileLoader]) -> None:
        """Register a PileLoader class with the given key."""
        if issubclass(loader, PileLoader):
            cls._loaders[key] = loader()
        elif isinstance(loader, PileLoader):
            cls._loaders[key] = loader
        else:
            raise LionTypeError(
                "Invalid loader class provided. Must be a subclass of "
                "PileLoader."
            )

    @classmethod
    def get(cls, key: str) -> type[PileLoader]:
        if key not in cls._loaders:
            raise KeyError(f"No loader registered for key: {key}")
        return cls._loaders[key]

    @classmethod
    def load_from(cls, obj: Any, key: str | None = None) -> list[dict]:
        if key:
            loader = cls.get(key)
            if loader.can_load(obj):
                return loader.from_obj(obj)
            raise LionValueError(f"Loader {key} cannot load the provided data")

        for loader in cls._loaders.values():
            if loader.can_load(obj):
                return loader.from_obj(obj)

        raise LionTypeError(
            f"No suitable loader found for the provided data type: {type(obj)}"
        )



# import os
# from datetime import datetime, timezone
# from pathlib import Path
# from unittest.mock import MagicMock, patch

# import pytest

# from lionabc import Observable
# from lionabc.exceptions import LionIDError
# from lion_core.setting import LionIDConfig
# from lion_core.sys_utils import SysUtil


# class TestSysUtilTime:
#     def test_time_timestamp(self):
#         result = SysUtil.time(type_="timestamp")
#         assert isinstance(result, float)
#         assert result > 0

#     def test_time_datetime(self):
#         result = SysUtil.time(type_="datetime")
#         assert isinstance(result, datetime)
#         assert result.tzinfo == timezone.utc

#     def test_time_iso(self):
#         result = SysUtil.time(type_="iso")
#         assert isinstance(result, str)
#         assert "T" in result
#         assert result.endswith("Z") or result.endswith("+00:00")

#     def test_time_custom(self):
#         custom_format = "%Y-%m-%d %H:%M:%S"
#         result = SysUtil.time(type_="custom", custom_format=custom_format)
#         assert isinstance(result, str)
#         assert datetime.strptime(result, custom_format)

#     def test_time_invalid_type(self):
#         with pytest.raises(ValueError):
#             SysUtil.time(type_="invalid")

#     def test_time_custom_without_format(self):
#         with pytest.raises(ValueError):
#             SysUtil.time(type_="custom")

#     def test_time_custom_sep(self):
#         result = SysUtil.time(
#             type_="custom", custom_format="%Y-%m-%d %H:%M:%S", custom_sep="_"
#         )
#         assert "_" in result
#         assert "-" not in result

#     def test_time_timestamp(self):
#         result = SysUtil.time(type_="timestamp")
#         assert isinstance(result, float)
#         assert result > 0

#     def test_time_datetime(self):
#         result = SysUtil.time(type_="datetime")
#         assert isinstance(result, datetime)
#         assert result.tzinfo == timezone.utc

#     def test_time_iso(self):
#         result = SysUtil.time(type_="iso")
#         assert isinstance(result, str)
#         assert "T" in result
#         assert result.endswith("Z") or result.endswith("+00:00")

#     def test_time_custom(self):
#         custom_format = "%Y-%m-%d %H:%M:%S"
#         result = SysUtil.time(type_="custom", custom_format=custom_format)
#         assert isinstance(result, str)
#         assert datetime.strptime(result, custom_format)

#     def test_time_invalid_type(self):
#         with pytest.raises(ValueError):
#             SysUtil.time(type_="invalid")

#     def test_time_custom_without_format(self):
#         with pytest.raises(ValueError):
#             SysUtil.time(type_="custom")


# class TestSysUtilCopy:
#     @pytest.fixture
#     def temp_directory(self, tmp_path):
#         base_dir = tmp_path / "test_dir"
#         base_dir.mkdir()
#         (base_dir / "file1.txt").write_text("content1")
#         (base_dir / "file2.py").write_text("content2")
#         sub_dir = base_dir / "subdir"
#         sub_dir.mkdir()
#         (sub_dir / "file3.txt").write_text("content3")
#         return base_dir

#     def test_copy_file(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "copy_of_file1.txt"
#         SysUtil.copy_file(src, dest)
#         assert dest.exists()
#         assert dest.read_text() == "content1"

#     def test_copy_file_non_existent_source(self, temp_directory):
#         src = temp_directory / "non_existent.txt"
#         dest = temp_directory / "copy.txt"
#         with pytest.raises(FileNotFoundError):
#             SysUtil.copy_file(src, dest)

#     def test_copy_file_permission_error(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "no_permission.txt"
#         with patch("lion_core.sys_utils.copy2", side_effect=PermissionError):
#             with pytest.raises(PermissionError):
#                 SysUtil.copy_file(src, dest)

#     def test_copy_file_os_error(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "os_error.txt"
#         with patch("lion_core.sys_utils.copy2", side_effect=OSError):
#             with pytest.raises(OSError):
#                 SysUtil.copy_file(src, dest)

#     def test_copy_file_overwrite(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "file2.py"
#         SysUtil.copy_file(src, dest)
#         assert dest.read_text() == "content1"

#     def test_copy_file_to_new_directory(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "new_dir" / "copy_of_file1.txt"
#         SysUtil.copy_file(src, dest)
#         assert dest.exists()
#         assert dest.read_text() == "content1"

#     @pytest.mark.skipif(
#         os.name == "nt", reason="Symlinks might not be available on Windows"
#     )
#     def test_copy_file_symlink(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         link = temp_directory / "link_to_file1.txt"
#         link.symlink_to(src)
#         dest = temp_directory / "copy_of_link.txt"
#         SysUtil.copy_file(link, dest)
#         assert dest.exists()
#         assert not dest.is_symlink()
#         assert dest.read_text() == "content1"


# import os
# import re
# from datetime import datetime
# from pathlib import Path
# from unittest.mock import MagicMock, patch

# import pytest

# from lion_core.sys_utils import SysUtil


# class TestSysUtilPathOperations:
#     @pytest.fixture
#     def temp_dir(self, tmp_path):
#         return tmp_path

#     def test_list_files_all(self, temp_dir):
#         (temp_dir / "file1.txt").touch()
#         (temp_dir / "file2.py").touch()
#         subdir = temp_dir / "subdir"
#         subdir.mkdir()
#         (subdir / "file3.txt").touch()

#         files = SysUtil.list_files(temp_dir)
#         assert len(files) == 3
#         assert {f.name for f in files} == {
#             "file1.txt",
#             "file2.py",
#             "file3.txt",
#         }

#     def test_list_files_with_extension(self, temp_dir):
#         (temp_dir / "file1.txt").touch()
#         (temp_dir / "file2.py").touch()
#         (temp_dir / "subdir").mkdir()
#         (temp_dir / "subdir" / "file3.txt").touch()

#         files = SysUtil.list_files(temp_dir, "txt")
#         assert len(files) == 2
#         assert {f.name for f in files} == {"file1.txt", "file3.txt"}

#     def test_list_files_empty_directory(self, temp_dir):
#         files = SysUtil.list_files(temp_dir)
#         assert len(files) == 0

#     def test_list_files_non_existent_directory(self):
#         with pytest.raises(NotADirectoryError):
#             SysUtil.list_files("/non/existent/directory")

#     def test_list_files_file_as_input(self, temp_dir):
#         file_path = temp_dir / "test.txt"
#         file_path.touch()
#         with pytest.raises(NotADirectoryError):
#             SysUtil.list_files(file_path)

#     def test_split_path(self):
#         path = Path("/home/user/documents/file.txt")
#         dir_path, filename = SysUtil.split_path(path)
#         assert dir_path == Path("/home/user/documents")
#         assert filename == "file.txt"

#     def test_split_path_just_filename(self):
#         path = "file.txt"
#         dir_path, filename = SysUtil.split_path(path)
#         assert dir_path == Path(".")
#         assert filename == "file.txt"

#     def test_split_path_root(self):
#         path = "/"
#         dir_path, filename = SysUtil.split_path(path)
#         assert dir_path == Path("/")
#         assert filename == ""

#     def test_get_file_size_file(self, temp_dir):
#         file_path = temp_dir / "test.txt"
#         file_path.write_text("content")
#         size = SysUtil.get_file_size(file_path)
#         assert size == 7  # "content" is 7 bytes

#     def test_get_file_size_directory(self, temp_dir):
#         (temp_dir / "file1.txt").write_text("content1")
#         (temp_dir / "file2.txt").write_text("content2")
#         size = SysUtil.get_file_size(temp_dir)
#         assert size == 16  # "content1" + "content2" = 8 + 8 = 16 bytes

#     def test_get_file_size_non_existent(self):
#         with pytest.raises(FileNotFoundError):
#             SysUtil.get_file_size("/non/existent/path")

#     def test_get_file_size_permission_error(self, temp_dir):
#         with patch.object(Path, "stat", side_effect=PermissionError):
#             with pytest.raises(PermissionError):
#                 SysUtil.get_file_size(temp_dir)

#     def test_get_file_size_empty_dir(self, temp_dir):
#         assert SysUtil.get_file_size(temp_dir) == 0


# class TestSysUtilPathCreation:
#     @pytest.fixture
#     def temp_dir(self, tmp_path):
#         return tmp_path

#     def test_create_path_basic(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt")
#         assert path.parent == temp_dir
#         assert path.name == "test.txt"
#         assert (
#             not path.exists()
#         )  # File should not be created, only path returned

#     def test_create_path_with_timestamp(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt", timestamp=True)
#         assert path.parent == temp_dir
#         assert re.match(r"test_\d{14}\.txt", path.name)

#     def test_create_path_with_timestamp_prefix(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir, "test.txt", timestamp=True, time_prefix=True
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"\d{14}_test\.txt", path.name)

#     def test_create_path_custom_timestamp(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir, "test.txt", timestamp=True, timestamp_format="%Y%m%d"
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"test_\d{8}\.txt", path.name)

#     def test_create_path_with_random_hash(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt", random_hash_digits=5)
#         assert path.parent == temp_dir
#         assert re.match(r"test-[a-f0-9]{5}\.txt", path.name)

#     def test_create_path_file_exists(self, temp_dir):
#         existing_file = temp_dir / "existing.txt"
#         existing_file.touch()
#         path = SysUtil.create_path(
#             temp_dir, "existing.txt", file_exist_ok=True
#         )
#         assert path == existing_file
#         assert path.exists()

#     def test_create_path_file_exists_not_ok(self, temp_dir):
#         existing_file = temp_dir / "existing.txt"
#         existing_file.touch()
#         with pytest.raises(FileExistsError):
#             SysUtil.create_path(temp_dir, "existing.txt", file_exist_ok=False)

#     def test_create_path_nested_directory(self, temp_dir):
#         nested_dir = temp_dir / "nested" / "dir"
#         path = SysUtil.create_path(nested_dir, "test.txt")
#         assert path.parent == nested_dir
#         assert path.name == "test.txt"
#         assert nested_dir.exists()  # Directory should be created
#         assert not path.exists()  # File should not be created

#     def test_create_path_invalid_filename(self, temp_dir):
#         with pytest.raises(ValueError):
#             SysUtil.create_path(temp_dir, "invalid/filename.txt")

#     def test_create_path_all_options(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir,
#             "test.txt",
#             timestamp=True,
#             time_prefix=True,
#             random_hash_digits=5,
#             timestamp_format="%Y%m%d",
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"\d{8}_test-[a-f0-9]{5}\.txt", path.name)


# class TestSysUtilFileOperations:
#     @pytest.fixture
#     def temp_dir(self, tmp_path):
#         return tmp_path

#     def test_save_to_file(self, temp_dir):
#         text = "Hello, World!"
#         file_path = SysUtil.save_to_file(
#             text, temp_dir, "test.txt", timestamp=False, random_hash_digits=0
#         )
#         assert file_path.exists()
#         assert file_path.read_text() == text

#     def test_save_to_file_with_timestamp(self, temp_dir):
#         text = "Hello, World!"
#         file_path = SysUtil.save_to_file(
#             text, temp_dir, "test.txt", timestamp=True
#         )
#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert re.match(r"test_\d{14}\.txt", file_path.name)

#     def test_save_to_file_with_random_hash(self, temp_dir):
#         text = "Hello, World!"
#         file_path = SysUtil.save_to_file(
#             text, temp_dir, "test.txt", timestamp=False, random_hash_digits=6
#         )
#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert re.match(r"test-[a-f0-9]{6}\.txt", file_path.name)

#     def test_read_file(self, temp_dir):
#         text = "Hello, World!"
#         file_path = temp_dir / "test.txt"
#         file_path.write_text(text)
#         result = SysUtil.read_file(file_path)
#         assert result == text

#     def test_read_file_not_found(self, temp_dir):
#         non_existent_file = temp_dir / "non_existent.txt"
#         with pytest.raises(FileNotFoundError):
#             SysUtil.read_file(non_existent_file)

#     def test_save_to_file_nested_directory(self, temp_dir):
#         text = "Hello, World!"
#         nested_dir = temp_dir / "nested" / "dir"
#         file_path = SysUtil.save_to_file(text, nested_dir, "test.txt")
#         assert file_path.exists()
#         assert file_path.read_text() == text

#     def test_save_to_file_overwrite(self, temp_dir):
#         file_path = temp_dir / "test.txt"

#         SysUtil.save_to_file(
#             "Original", temp_dir, "test.txt", file_exist_ok=True
#         )
#         SysUtil.save_to_file(
#             "Updated", temp_dir, "test.txt", file_exist_ok=True
#         )
#         assert file_path.read_text() == "Updated"

#     @pytest.mark.parametrize(
#         "filename",
#         [
#             "normal.txt",
#             "with spaces.txt",
#             "with-dashes.txt",
#             "with_underscores.txt",
#             "with.multiple.dots.txt",
#             "တောင်ကြီးမြို့.txt",  # Non-ASCII characters
#             "🐍.py",  # Emoji
#         ],
#     )
#     def test_file_operations_with_special_filenames(self, temp_dir, filename):
#         content = "Test content"
#         file_path = SysUtil.save_to_file(content, temp_dir, filename)
#         assert file_path.exists()
#         assert SysUtil.read_file(file_path) == content

#     @pytest.mark.skipif(
#         os.name == "nt", reason="Symlinks might not be available on Windows"
#     )
#     def test_file_operations_with_symlink(self, temp_dir):
#         original = temp_dir / "original.txt"
#         SysUtil.save_to_file("Original content", temp_dir, "original.txt")
#         symlink = temp_dir / "symlink.txt"
#         symlink.symlink_to(original)

#         # Test reading through symlink
#         content = SysUtil.read_file(symlink)
#         assert content == "Original content"

#         # Test file size through symlink
#         size = SysUtil.get_file_size(symlink)
#         assert size == len("Original content")


# import importlib
# import subprocess
# from unittest.mock import MagicMock, patch

# import pytest

# from lion_core.sys_utils import SysUtil


# class TestSysUtilImportMethods:

#     @patch("lion_core.sys_utils.SysUtil.import_module")
#     @patch("lion_core.sys_utils._run_pip_command")
#     def test_install_import_success(self, mock_run_pip, mock_import):
#         mock_import.side_effect = [ImportError, MagicMock()]

#         result = SysUtil.install_import("test_package")

#         assert mock_run_pip.called
#         assert mock_import.call_count == 2
#         assert result is not None

#     @patch("lion_core.sys_utils.SysUtil.import_module")
#     @patch("lion_core.sys_utils._run_pip_command")
#     def test_install_import_failure(self, mock_run_pip, mock_import):
#         mock_import.side_effect = ImportError
#         mock_run_pip.side_effect = subprocess.CalledProcessError(1, "pip")

#         with pytest.raises(ImportError):
#             SysUtil.install_import("test_package")

#     @patch("builtins.__import__")
#     def test_import_module_success(self, mock_import):
#         mock_module = MagicMock()
#         mock_module.TestClass = "TestClass"
#         mock_import.return_value = mock_module

#         result = SysUtil.import_module(
#             "test_package", "test_module", "TestClass"
#         )

#         assert result == "TestClass"
#         mock_import.assert_called_with(
#             "test_package.test_module", fromlist=["TestClass"]
#         )

#     @patch("builtins.__import__")
#     def test_import_module_without_module_name(self, mock_import):
#         mock_module = MagicMock()
#         mock_import.return_value = mock_module

#         result = SysUtil.import_module("test_package")

#         assert result == mock_module
#         mock_import.assert_called_with("test_package")

#     @patch("builtins.__import__")
#     def test_import_module_with_multiple_import_names(self, mock_import):
#         mock_module = MagicMock()
#         mock_module.Class1 = "Class1"
#         mock_module.Class2 = "Class2"
#         mock_import.return_value = mock_module

#         result = SysUtil.import_module(
#             "test_package", "test_module", ["Class1", "Class2"]
#         )

#         assert result == ["Class1", "Class2"]
#         mock_import.assert_called_with(
#             "test_package.test_module", fromlist=["Class1", "Class2"]
#         )

#     def test_import_module_failure(self):
#         with pytest.raises(ImportError):
#             SysUtil.import_module("non_existent_package")

#     @patch("importlib.util.find_spec")
#     def test_is_package_installed_true(self, mock_find_spec):
#         mock_find_spec.return_value = MagicMock()

#         assert SysUtil.is_package_installed("existing_package") is True

#     @patch("importlib.util.find_spec")
#     def test_is_package_installed_false(self, mock_find_spec):
#         mock_find_spec.return_value = None

#         assert SysUtil.is_package_installed("non_existent_package") is False

#     @patch("lion_core.sys_utils.SysUtil.is_package_installed")
#     @patch("lion_core.sys_utils.SysUtil.install_import")
#     @patch("lion_core.sys_utils.SysUtil.import_module")
#     def test_check_import_installed(
#         self, mock_import, mock_install, mock_is_installed
#     ):
#         mock_is_installed.return_value = True
#         mock_import.return_value = MagicMock()

#         result = SysUtil.check_import("existing_package")

#         assert result is not None
#         assert not mock_install.called

#     @patch("lion_core.sys_utils.SysUtil.is_package_installed")
#     @patch("lion_core.sys_utils.SysUtil.install_import")
#     def test_check_import_not_installed_attempt_install(
#         self, mock_install, mock_is_installed
#     ):
#         mock_is_installed.return_value = False
#         mock_install.return_value = MagicMock()

#         result = SysUtil.check_import("new_package", attempt_install=True)

#         assert result is not None
#         assert mock_install.called

#     @patch("lion_core.sys_utils.SysUtil.is_package_installed")
#     def test_check_import_not_installed_no_attempt(self, mock_is_installed):
#         mock_is_installed.return_value = False

#         with pytest.raises(ImportError):
#             SysUtil.check_import("new_package", attempt_install=False)

#     @patch("lion_core.sys_utils.SysUtil.is_package_installed")
#     @patch("lion_core.sys_utils.SysUtil.install_import")
#     def test_check_import_with_custom_error_message(
#         self, mock_install, mock_is_installed
#     ):
#         mock_is_installed.return_value = False

#         with pytest.raises(ImportError, match="Custom error message"):
#             SysUtil.check_import(
#                 "new_package",
#                 attempt_install=False,
#                 error_message="Custom error message",
#             )

#     @patch("importlib.metadata.distributions")
#     def test_list_installed_packages(self, mock_distributions):
#         mock_dist1 = MagicMock()
#         mock_dist1.metadata = {"Name": "package1"}
#         mock_dist2 = MagicMock()
#         mock_dist2.metadata = {"Name": "package2"}
#         mock_distributions.return_value = [mock_dist1, mock_dist2]

#         result = SysUtil.list_installed_packages()

#         assert result == ["package1", "package2"]

#     @patch("importlib.metadata.distributions")
#     def test_list_installed_packages_error(self, mock_distributions):
#         mock_distributions.side_effect = Exception("Test error")

#         result = SysUtil.list_installed_packages()

#         assert result == []

#     @patch("lion_core.sys_utils._run_pip_command")
#     def test_uninstall_package_success(self, mock_run_pip):
#         SysUtil.uninstall_package("test_package")
#         mock_run_pip.assert_called_with(["uninstall", "test_package", "-y"])

#     @patch("lion_core.sys_utils._run_pip_command")
#     def test_uninstall_package_failure(self, mock_run_pip):
#         mock_run_pip.side_effect = subprocess.CalledProcessError(1, "pip")

#         with pytest.raises(subprocess.CalledProcessError):
#             SysUtil.uninstall_package("test_package")

#     @patch("lion_core.sys_utils._run_pip_command")
#     def test_update_package_success(self, mock_run_pip):
#         SysUtil.update_package("test_package")
#         mock_run_pip.assert_called_with(
#             ["install", "--upgrade", "test_package"]
#         )

#     @patch("lion_core.sys_utils._run_pip_command")
#     def test_update_package_failure(self, mock_run_pip):
#         mock_run_pip.side_effect = subprocess.CalledProcessError(1, "pip")

#         with pytest.raises(subprocess.CalledProcessError):
#             SysUtil.update_package("test_package")


# import os
# from pathlib import Path
# from unittest.mock import MagicMock, patch

# import pytest

# from lion_core.sys_utils import SysUtil


# class TestSysUtilAdditional:
#     @pytest.mark.parametrize(
#         "machine,expected",
#         [
#             ("arm64", "arm64"),
#             ("aarch64", "arm64"),
#             ("x86_64", "x86_64"),
#             ("amd64", "x86_64"),
#             ("i386", "i386"),
#             ("powerpc", "powerpc"),
#         ],
#     )
#     def test_get_cpu_architecture(self, machine, expected):
#         with patch("platform.machine", return_value=machine):
#             assert SysUtil.get_cpu_architecture() == expected

#     def test_get_cpu_architecture_unknown(self):
#         with patch("platform.machine", return_value="unknown_arch"):
#             assert SysUtil.get_cpu_architecture() == "unknown_arch"

#     @pytest.mark.skipif(
#         os.name == "nt", reason="Test specific to non-Windows systems"
#     )
#     def test_get_cpu_architecture_apple_silicon(self):
#         with patch("platform.machine", return_value="arm64"):
#             assert SysUtil.get_cpu_architecture() == "arm64"

#     def test_create_path_file_collision(self, tmp_path):
#         # Create the initial file
#         initial_path = SysUtil.create_path(tmp_path, "test.txt")
#         initial_path.touch()

#         # Attempt to create the same file again should raise FileExistsError
#         with pytest.raises(FileExistsError):
#             SysUtil.create_path(tmp_path, "test.txt", file_exist_ok=False)

#         # Creating with file_exist_ok=True should return the existing path
#         existing_path = SysUtil.create_path(
#             tmp_path, "test.txt", file_exist_ok=True
#         )
#         assert existing_path == initial_path

#         # Creating a file with a different name should succeed
#         new_path = SysUtil.create_path(
#             tmp_path, "new_test.txt", file_exist_ok=False
#         )
#         assert new_path != initial_path
#         assert new_path.name == "new_test.txt"

#     def test_get_path_kwargs_simple(self):
#         kwargs = SysUtil._get_path_kwargs("/path/to/file.txt", "txt")
#         assert kwargs["directory"] == Path("/path/to")
#         assert kwargs["filename"] == "file.txt"

#     def test_get_path_kwargs_no_extension(self):
#         kwargs = SysUtil._get_path_kwargs("/path/to/file", "txt")
#         assert kwargs["directory"] == Path("/path/to/file")
#         assert kwargs["filename"] == "new_file.txt"

#     def test_get_path_kwargs_directory_only(self):
#         kwargs = SysUtil._get_path_kwargs("/path/to/dir", "txt")
#         assert kwargs["directory"] == Path("/path/to/dir")
#         assert kwargs["filename"] == "new_file.txt"

#     @pytest.mark.parametrize(
#         "persist_path,postfix,expected",
#         [
#             (
#                 "/path/to/file.csv",
#                 "csv",
#                 {"directory": Path("/path/to"), "filename": "file.csv"},
#             ),
#             (
#                 "/path/to/dir",
#                 "txt",
#                 {
#                     "directory": Path("/path/to/dir"),
#                     "filename": "new_file.txt",
#                 },
#             ),
#             (
#                 "file.json",
#                 "json",
#                 {"directory": Path("."), "filename": "file.json"},
#             ),
#         ],
#     )
#     def test_get_path_kwargs_various_inputs(
#         self, persist_path, postfix, expected
#     ):
#         kwargs = SysUtil._get_path_kwargs(persist_path, postfix)
#         assert kwargs["directory"] == expected["directory"]
#         assert kwargs["filename"] == expected["filename"]

#     def test_large_directory(self, tmp_path):
#         large_dir = tmp_path / "large_dir"
#         large_dir.mkdir()
#         for i in range(10000):
#             (large_dir / f"file_{i}.txt").write_text("content")

#         files = SysUtil.list_files(large_dir)
#         assert len(files) == 10000

#         size = SysUtil.get_file_size(large_dir)
#         assert size == 7 * 10000  # "content" is 7 bytes, 10000 files


# # File: tests/test_sys_utils.py


# import os
# import re
# from datetime import datetime, timezone
# from pathlib import Path
# from shutil import rmtree
# from unittest.mock import patch

# import pytest

# from lionabc import Observable
# from lionabc.exceptions import LionIDError
# from lion_core.setting import LionIDConfig
# from lion_core.sys_utils import SysUtil


# class TestSysUtil:

#     def test_id_default(self):
#         id_ = SysUtil.id()
#         assert isinstance(id_, str)
#         assert len(id_) == 48  # 42 chars + prefix "ln" + 4 hyphens

#     def test_id_custom_length(self):
#         id_ = SysUtil.id(n=20)
#         assert len(id_) == 26  # 20 chars + prefix "ln" + 4 hyphens

#     def test_id_no_hyphens(self):
#         id_ = SysUtil.id(random_hyphen=False)
#         assert "-" not in id_

#     def test_id_custom_prefix_postfix(self):
#         id_ = SysUtil.id(prefix="test_", postfix="_end")
#         assert id_.startswith("test_")
#         assert id_.endswith("_end")

#     def test_get_id_valid(self):
#         class MockObservable(Observable):
#             def __init__(self, ln_id):
#                 self.ln_id = ln_id

#         valid_id = SysUtil.id()
#         mock_obj = MockObservable(valid_id)
#         assert SysUtil.get_id(mock_obj) == valid_id
#         assert SysUtil.get_id(valid_id) == valid_id

#     def test_get_id_invalid(self):
#         with pytest.raises(LionIDError):
#             SysUtil.get_id("invalid_id")

#     def test_get_id_sequence(self):
#         valid_id = SysUtil.id()
#         assert SysUtil.get_id([valid_id]) == valid_id

#     def test_get_id_custom_config(self):
#         custom_config = LionIDConfig(
#             prefix="custom-",
#             n=20,
#             num_hyphens=2,
#             random_hyphen=True,
#             hyphen_start_index=6,
#             hyphen_end_index=-6,
#         )
#         valid_id = "custom-a32eac2f22a-62-a45092f"
#         assert SysUtil.get_id(valid_id, custom_config) == valid_id

#     def test_is_id_valid(self):
#         valid_id = "ln8d7274-45a36cc0-8b89af8b-c5-154c41996cf228d8a5"
#         assert SysUtil.is_id(valid_id) == True

#     def test_is_id_invalid(self):
#         invalid_id = "invalid_id"
#         assert SysUtil.is_id(invalid_id) == False

#     def test_is_id_custom_config(self):
#         custom_config = LionIDConfig(
#             prefix="custom-",
#             n=20,
#             num_hyphens=2,
#             random_hyphen=True,
#             hyphen_start_index=6,
#             hyphen_end_index=-6,
#         )
#         valid_id = "custom-a32eac2f22a-62-a45092f"
#         assert SysUtil.is_id(valid_id, custom_config) == True

#     # Additional tests for edge cases
#     def test_get_id_32_char(self):
#         valid_id = "a" * 32
#         assert SysUtil.get_id(valid_id) == valid_id

#     def test_get_id_invalid_type(self):
#         with pytest.raises(LionIDError):
#             SysUtil.get_id(123)

#     def test_is_id_edge_cases(self):
#         assert SysUtil.is_id("a" * 32) == True
#         assert SysUtil.is_id(123) == False


# import os
# import tempfile
# from datetime import datetime
# from pathlib import Path

# import pytest

# from lion_core.sys_utils import SysUtil


# class TestSysUtilPathMethods:

#     @pytest.fixture
#     def temp_dir(self):
#         with tempfile.TemporaryDirectory() as tmpdirname:
#             yield Path(tmpdirname)

#     def test_clear_path_recursive(self, temp_dir):
#         subdir = temp_dir / "subdir"
#         subdir.mkdir()
#         (subdir / "file.txt").touch()
#         SysUtil.clear_path(temp_dir, recursive=True)
#         assert len(list(temp_dir.iterdir())) == 0

#     def test_create_path_basic(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt")
#         assert path.parent == temp_dir
#         assert path.name == "test.txt"

#     def test_create_path_with_timestamp(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt", timestamp=True)
#         assert path.parent == temp_dir
#         assert re.match(r"test_\d{14}\.txt", path.name)

#     def test_create_path_custom_timestamp(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir, "test.txt", timestamp=True, timestamp_format="%Y%m%d"
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"test_\d{8}\.txt", path.name)

#     def test_create_path_file_exists(self, temp_dir):
#         existing_file = temp_dir / "existing.txt"
#         existing_file.touch()
#         path = SysUtil.create_path(
#             temp_dir, "existing.txt", file_exist_ok=True
#         )
#         assert path == existing_file

#     def test_get_path_kwargs_simple(self):
#         kwargs = SysUtil._get_path_kwargs("/path/to/file.txt", "txt")
#         assert kwargs["directory"] == Path("/path/to")
#         assert kwargs["filename"] == "file.txt"

#     def test_get_path_kwargs_no_extension(self):
#         kwargs = SysUtil._get_path_kwargs("/path/to/file", "txt")
#         assert kwargs["directory"] == Path("/path/to/file")
#         assert kwargs["filename"] == "new_file.txt"

#     def test_get_path_kwargs_directory_only(self):
#         kwargs = SysUtil._get_path_kwargs("/path/to/dir", "txt")
#         assert kwargs["directory"] == Path("/path/to/dir")
#         assert kwargs["filename"] == "new_file.txt"

#     @pytest.mark.parametrize(
#         "persist_path,postfix,expected",
#         [
#             (
#                 "/path/to/file.csv",
#                 "csv",
#                 {"directory": Path("/path/to"), "filename": "file.csv"},
#             ),
#             (
#                 "/path/to/dir",
#                 "txt",
#                 {
#                     "directory": Path("/path/to/dir"),
#                     "filename": "new_file.txt",
#                 },
#             ),
#             (
#                 "file.json",
#                 "json",
#                 {"directory": Path("."), "filename": "file.json"},
#             ),
#         ],
#     )
#     def test_get_path_kwargs_various_inputs(
#         self, persist_path, postfix, expected
#     ):
#         kwargs = SysUtil._get_path_kwargs(persist_path, postfix)
#         assert kwargs["directory"] == expected["directory"]
#         assert kwargs["filename"] == expected["filename"]


# import re
# from datetime import datetime
# from pathlib import Path

# import pytest

# from lion_core.sys_utils import SysUtil


# class TestSysUtilCreatePath:
#     @pytest.fixture
#     def temp_dir(self, tmp_path):
#         return tmp_path

#     def test_create_path_basic(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt")
#         assert path.parent == temp_dir
#         assert path.name == "test.txt"
#         assert (
#             not path.exists()
#         )  # File should not be created, only path returned

#     def test_create_path_with_timestamp(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt", timestamp=True)
#         assert path.parent == temp_dir
#         assert re.match(r"test_\d{14}\.txt", path.name)

#     def test_create_path_with_timestamp_prefix(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir, "test.txt", timestamp=True, time_prefix=True
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"\d{14}_test\.txt", path.name)

#     def test_create_path_custom_timestamp(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir, "test.txt", timestamp=True, timestamp_format="%Y%m%d"
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"test_\d{8}\.txt", path.name)

#     def test_create_path_with_random_hash(self, temp_dir):
#         path = SysUtil.create_path(temp_dir, "test.txt", random_hash_digits=5)
#         assert path.parent == temp_dir
#         assert re.match(r"test-[a-f0-9]{5}\.txt", path.name)

#     def test_create_path_file_exists(self, temp_dir):
#         existing_file = temp_dir / "existing.txt"
#         existing_file.touch()
#         path = SysUtil.create_path(
#             temp_dir, "existing.txt", file_exist_ok=True
#         )
#         assert path == existing_file
#         assert path.exists()

#     def test_create_path_file_exists_not_ok(self, temp_dir):
#         existing_file = temp_dir / "existing.txt"
#         existing_file.touch()
#         with pytest.raises(FileExistsError):
#             path = SysUtil.create_path(
#                 temp_dir, "existing.txt", file_exist_ok=False
#             )

#     def test_create_path_nested_directory(self, temp_dir):
#         nested_dir = temp_dir / "nested" / "dir"
#         path = SysUtil.create_path(nested_dir, "test.txt")
#         assert path.parent == nested_dir
#         assert path.name == "test.txt"
#         assert nested_dir.exists()  # Directory should be created
#         assert not path.exists()  # File should not be created

#     def test_create_path_invalid_filename(self, temp_dir):
#         with pytest.raises(ValueError):
#             SysUtil.create_path(temp_dir, "invalid/filename.txt")

#     def test_create_path_all_options(self, temp_dir):
#         path = SysUtil.create_path(
#             temp_dir,
#             "test.txt",
#             timestamp=True,
#             time_prefix=True,
#             random_hash_digits=5,
#             timestamp_format="%Y%m%d",
#         )
#         assert path.parent == temp_dir
#         assert re.match(r"\d{8}_test-[a-f0-9]{5}\.txt", path.name)

#     def test_create_path_file_collision(self, temp_dir):
#         # Create the initial file
#         initial_path = SysUtil.create_path(temp_dir, "test.txt")
#         initial_path.touch()

#         # Attempt to create the same file again should raise FileExistsError
#         with pytest.raises(FileExistsError):
#             SysUtil.create_path(temp_dir, "test.txt", file_exist_ok=False)

#         # Creating with file_exist_ok=True should return the existing path
#         existing_path = SysUtil.create_path(
#             temp_dir, "test.txt", file_exist_ok=True
#         )
#         assert existing_path == initial_path

#         # Creating a file with a different name should succeed
#         new_path = SysUtil.create_path(
#             temp_dir, "new_test.txt", file_exist_ok=False
#         )
#         assert new_path != initial_path
#         assert new_path.name == "new_test.txt"


# @pytest.fixture
# def temp_directory(tmp_path):
#     """Create a temporary directory with some files and subdirectories."""
#     base_dir = tmp_path / "test_dir"
#     base_dir.mkdir()
#     (base_dir / "file1.txt").write_text("content1")
#     (base_dir / "file2.py").write_text("content2")
#     sub_dir = base_dir / "subdir"
#     sub_dir.mkdir()
#     (sub_dir / "file3.txt").write_text("content3")
#     yield base_dir
#     rmtree(base_dir)


# class TestFileUtils:

#     def test_list_files_all(self, temp_directory):
#         files = SysUtil.list_files(temp_directory)
#         assert len(files) == 3
#         assert {f.name for f in files} == {
#             "file1.txt",
#             "file2.py",
#             "file3.txt",
#         }

#     def test_list_files_with_extension(self, temp_directory):
#         files = SysUtil.list_files(temp_directory, "txt")
#         assert len(files) == 2
#         assert {f.name for f in files} == {"file1.txt", "file3.txt"}

#     def test_list_files_empty_directory(self, tmp_path):
#         empty_dir = tmp_path / "empty"
#         empty_dir.mkdir()
#         files = SysUtil.list_files(empty_dir)
#         assert len(files) == 0

#     def test_list_files_non_existent_directory(self):
#         with pytest.raises(NotADirectoryError):
#             SysUtil.list_files("/non/existent/directory")

#     def test_list_files_file_as_input(self, temp_directory):
#         with pytest.raises(NotADirectoryError):
#             SysUtil.list_files(temp_directory / "file1.txt")

#     def test_split_path(self):
#         path = Path("/home/user/documents/file.txt")
#         dir_path, filename = SysUtil.split_path(path)
#         assert dir_path == Path("/home/user/documents")
#         assert filename == "file.txt"

#     def test_split_path_just_filename(self):
#         path = "file.txt"
#         dir_path, filename = SysUtil.split_path(path)
#         assert dir_path == Path(".")
#         assert filename == "file.txt"

#     def test_copy_file(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "copy_of_file1.txt"
#         SysUtil.copy_file(src, dest)
#         assert dest.exists()
#         assert dest.read_text() == "content1"

#     def test_copy_file_non_existent_source(self, temp_directory):
#         src = temp_directory / "non_existent.txt"
#         dest = temp_directory / "copy.txt"
#         with pytest.raises(FileNotFoundError):
#             SysUtil.copy_file(src, dest)

#     def test_copy_file_permission_error(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "no_permission.txt"
#         with patch("lion_core.sys_utils.copy2", side_effect=PermissionError):
#             with pytest.raises(PermissionError):
#                 SysUtil.copy_file(src, dest)

#     def test_copy_file_os_error(self, temp_directory):
#         src = temp_directory / "file1.txt"
#         dest = temp_directory / "os_error.txt"
#         with patch("lion_core.sys_utils.copy2", side_effect=OSError):
#             with pytest.raises(OSError):
#                 SysUtil.copy_file(src, dest)

#     def test_get_file_size_file(self, temp_directory):
#         file_path = temp_directory / "file1.txt"
#         size = SysUtil.get_file_size(file_path)
#         assert size == 8  # "content1" is 8 bytes

#     def test_get_file_size_directory(self, temp_directory):
#         size = SysUtil.get_file_size(temp_directory)
#         assert (
#             size == 24
#         )  # Total content size: "content1" + "content2" + "content3" = 8 + 8 + 8 = 24 bytes

#     def test_get_file_size_non_existent(self):
#         with pytest.raises(FileNotFoundError):
#             SysUtil.get_file_size("/non/existent/path")

#     def test_get_file_size_permission_error(self, temp_directory):
#         with patch.object(Path, "stat", side_effect=PermissionError):
#             with pytest.raises(PermissionError):
#                 SysUtil.get_file_size(temp_directory)

#     @pytest.mark.parametrize(
#         "filename",
#         [
#             "normal.txt",
#             "with spaces.txt",
#             "with-dashes.txt",
#             "with_underscores.txt",
#             "with.multiple.dots.txt",
#             "တောင်ကြီးမြို့.txt",  # Non-ASCII characters
#             "🐍.py",  # Emoji
#         ],
#     )
#     def test_file_operations_with_special_filenames(
#         self, temp_directory, filename
#     ):
#         file_path = temp_directory / filename
#         file_path.write_text("content")

#         # Test list_files
#         files = SysUtil.list_files(temp_directory)
#         assert file_path in files

#         # Test split_path
#         dir_path, name = SysUtil.split_path(file_path)
#         assert dir_path == temp_directory
#         assert name == filename

#         # Test copy_file
#         dest_path = temp_directory / f"copy_of_{filename}"
#         SysUtil.copy_file(file_path, dest_path)
#         assert dest_path.exists()

#         # Test get_file_size
#         size = SysUtil.get_file_size(file_path)
#         assert size == 7  # "content" is 7 bytes

#     @pytest.mark.skipif(
#         os.name == "nt", reason="Symlinks might not be available on Windows"
#     )
#     def test_file_operations_with_symlink(self, temp_directory):
#         original = temp_directory / "original.txt"
#         original.write_text("original content")
#         symlink = temp_directory / "symlink.txt"
#         symlink.symlink_to(original)

#         # Test list_files
#         files = SysUtil.list_files(temp_directory)
#         assert symlink in files

#         # Test get_file_size
#         size = SysUtil.get_file_size(symlink)
#         assert size == len("original content")

#     def test_large_directory(self, tmp_path):
#         large_dir = tmp_path / "large_dir"
#         large_dir.mkdir()
#         for i in range(10000):
#             (large_dir / f"file_{i}.txt").write_text("content")

#         files = SysUtil.list_files(large_dir)
#         assert len(files) == 10000

#         size = SysUtil.get_file_size(large_dir)
#         assert size == 7 * 10000  # "content" is 7 bytes, 10000 files

#     def test_save_to_file(self, tmp_path):
#         text = "Hello, World!"
#         directory = tmp_path / "test_dir"
#         filename = "test_file.txt"

#         file_path = SysUtil.save_to_file(
#             text, directory, filename, timestamp=False, random_hash_digits=0
#         )

#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert file_path == directory / filename

#     def test_save_to_file_with_timestamp(self, tmp_path):
#         text = "Hello, World!"
#         directory = tmp_path / "test_dir"
#         filename = "test_file.txt"

#         file_path = SysUtil.save_to_file(
#             text, directory, filename, timestamp=True
#         )

#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert file_path.name.startswith("test_file_")
#         assert file_path.name.endswith(".txt")

#     def test_save_to_file_with_random_hash(self, tmp_path):
#         text = "Hello, World!"
#         directory = tmp_path / "test_dir"
#         filename = "test_file.txt"

#         file_path = SysUtil.save_to_file(
#             text, directory, filename, timestamp=False, random_hash_digits=6
#         )

#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert (
#             len(file_path.stem) == len("test_file") + 1 + 6
#         )  # name + hyphen + hash
#         assert file_path.name.endswith(".txt")

#     def test_read_file(self, tmp_path):
#         text = "Hello, World!"
#         file_path = tmp_path / "test_file.txt"
#         file_path.write_text(text)

#         result = SysUtil.read_file(file_path)

#         assert result == text

#     def test_read_file_not_found(self, tmp_path):
#         non_existent_file = tmp_path / "non_existent.txt"

#         with pytest.raises(FileNotFoundError):
#             SysUtil.read_file(non_existent_file)

#     @pytest.mark.parametrize(
#         "machine,expected",
#         [
#             ("arm64", "arm64"),
#             ("aarch64", "arm64"),
#             ("x86_64", "x86_64"),
#             ("amd64", "x86_64"),
#             ("i386", "i386"),
#             ("powerpc", "powerpc"),
#         ],
#     )
#     def test_get_cpu_architecture(self, machine, expected):
#         with patch("platform.machine", return_value=machine):
#             assert SysUtil.get_cpu_architecture() == expected


# # File: tests/test_sys_util.py

# import importlib
# import subprocess
# from pathlib import Path
# from unittest.mock import MagicMock, patch


# class TestSysUtil:

#     def test_save_to_file(self, tmp_path):
#         text = "Hello, World!"
#         directory = tmp_path / "test_dir"
#         filename = "test_file.txt"

#         file_path = SysUtil.save_to_file(
#             text, directory, filename, timestamp=False, random_hash_digits=0
#         )

#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert file_path == directory / filename

#     def test_save_to_file_with_timestamp(self, tmp_path):
#         text = "Hello, World!"
#         directory = tmp_path / "test_dir"
#         filename = "test_file.txt"

#         file_path = SysUtil.save_to_file(
#             text, directory, filename, timestamp=True
#         )

#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert file_path.name.startswith("test_file_")
#         assert file_path.name.endswith(".txt")

#     def test_save_to_file_with_random_hash(self, tmp_path):
#         text = "Hello, World!"
#         directory = tmp_path / "test_dir"
#         filename = "test_file.txt"

#         file_path = SysUtil.save_to_file(
#             text, directory, filename, timestamp=False, random_hash_digits=6
#         )

#         assert file_path.exists()
#         assert file_path.read_text() == text
#         assert (
#             len(file_path.stem) == len("test_file") + 1 + 6
#         )  # name + hyphen + hash
#         assert file_path.name.endswith(".txt")

#     def test_read_file(self, tmp_path):
#         text = "Hello, World!"
#         file_path = tmp_path / "test_file.txt"
#         file_path.write_text(text)

#         result = SysUtil.read_file(file_path)

#         assert result == text

#     def test_read_file_not_found(self, tmp_path):
#         non_existent_file = tmp_path / "non_existent.txt"

#         with pytest.raises(FileNotFoundError):
#             SysUtil.read_file(non_existent_file)

#     @pytest.mark.parametrize(
#         "machine,expected",
#         [
#             ("arm64", "arm64"),
#             ("aarch64", "arm64"),
#             ("x86_64", "x86_64"),
#             ("amd64", "x86_64"),
#             ("i386", "i386"),
#             ("powerpc", "powerpc"),
#         ],
#     )
#     def test_get_cpu_architecture(self, machine, expected):
#         with patch("platform.machine", return_value=machine):
#             assert SysUtil.get_cpu_architecture() == expected

#     @patch("lion_core.sys_utils.SysUtil.import_module")  # Updated patch path
#     @patch("lion_core.sys_utils._run_pip_command")  # Updated patch path
#     def test_install_import_success(self, mock_run_pip, mock_import):
#         mock_import.side_effect = [ImportError, MagicMock()]

#         result = SysUtil.install_import("test_package")

#         assert mock_run_pip.called
#         assert mock_import.call_count == 2
#         assert result is not None

#     @patch("lion_core.sys_utils.SysUtil.import_module")  # Updated patch path
#     @patch("lion_core.sys_utils._run_pip_command")  # Updated patch path
#     def test_install_import_failure(self, mock_run_pip, mock_import):
#         mock_import.side_effect = ImportError
#         mock_run_pip.side_effect = subprocess.CalledProcessError(1, "pip")

#         with pytest.raises(ImportError):
#             SysUtil.install_import("test_package")

#     @patch("builtins.__import__")
#     def test_import_module_success(self, mock_import):
#         mock_module = MagicMock()
#         mock_module.TestClass = "TestClass"
#         mock_import.return_value = mock_module

#         result = SysUtil.import_module(
#             "test_package", "test_module", "TestClass"
#         )

#         assert result == "TestClass"
#         mock_import.assert_called_with(
#             "test_package.test_module", fromlist=["TestClass"]
#         )

#     @patch("importlib.import_module")
#     def test_import_module_failure(self, mock_import):
#         mock_import.side_effect = ImportError

#         with pytest.raises(ImportError):
#             SysUtil.import_module("non_existent_package")

#     @patch("importlib.util.find_spec")
#     def test_is_package_installed_true(self, mock_find_spec):
#         mock_find_spec.return_value = MagicMock()

#         assert SysUtil.is_package_installed("existing_package") is True

#     @patch("importlib.util.find_spec")
#     def test_is_package_installed_false(self, mock_find_spec):
#         mock_find_spec.return_value = None

#         assert SysUtil.is_package_installed("non_existent_package") is False

#     @patch(
#         "lion_core.sys_utils.SysUtil.is_package_installed"
#     )  # Updated patch path
#     @patch("lion_core.sys_utils.SysUtil.install_import")  # Updated patch path
#     @patch("lion_core.sys_utils.SysUtil.import_module")  # Updated patch path
#     def test_check_import_installed(
#         self, mock_import, mock_install, mock_is_installed
#     ):
#         mock_is_installed.return_value = True
#         mock_import.return_value = MagicMock()

#         result = SysUtil.check_import("existing_package")

#         assert result is not None
#         assert not mock_install.called

#     @patch(
#         "lion_core.sys_utils.SysUtil.is_package_installed"
#     )  # Updated patch path
#     @patch("lion_core.sys_utils.SysUtil.install_import")  # Updated patch path
#     def test_check_import_not_installed_attempt_install(
#         self, mock_install, mock_is_installed
#     ):
#         mock_is_installed.return_value = False
#         mock_install.return_value = MagicMock()

#         result = SysUtil.check_import("new_package", attempt_install=True)

#         assert result is not None
#         assert mock_install.called

#     @patch(
#         "lion_core.sys_utils.SysUtil.is_package_installed"
#     )  # Updated patch path
#     def test_check_import_not_installed_no_attempt(self, mock_is_installed):
#         mock_is_installed.return_value = False

#         with pytest.raises(ImportError):
#             SysUtil.check_import("new_package", attempt_install=False)

#     @patch("importlib.metadata.distributions")
#     def test_list_installed_packages(self, mock_distributions):
#         mock_dist1 = MagicMock()
#         mock_dist1.metadata = {"Name": "package1"}
#         mock_dist2 = MagicMock()
#         mock_dist2.metadata = {"Name": "package2"}
#         mock_distributions.return_value = [mock_dist1, mock_dist2]

#         result = SysUtil.list_installed_packages()

#         assert result == ["package1", "package2"]

#     @patch("importlib.metadata.distributions")
#     def test_list_installed_packages_error(self, mock_distributions):
#         mock_distributions.side_effect = Exception("Test error")

#         result = SysUtil.list_installed_packages()

#         assert result == []


# # File: tests/test_sys_utils.py



import pytest

from lion_core.communication.message import MessageFlag, MessageRole
from lion_core.communication.system import (
    DEFAULT_SYSTEM,
    System,
    format_system_content,
)
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


# Tests for format_system_content function
def test_format_system_content_default():
    result = format_system_content(None, DEFAULT_SYSTEM)
    assert isinstance(result, Note)
    assert result.get("system_info") == DEFAULT_SYSTEM


def test_format_system_content_custom():
    custom_message = "Custom system message"
    result = format_system_content(None, custom_message)
    assert result.get("system_info") == custom_message


def test_format_system_content_with_datetime_flag():
    result = format_system_content(True, DEFAULT_SYSTEM)
    assert "System Date:" in result.get("system_info")


def test_format_system_content_with_datetime_string():
    custom_date = "2023-01-01 12:00"
    result = format_system_content(custom_date, DEFAULT_SYSTEM)
    assert custom_date in result.get("system_info")


# Tests for System class
def test_system_init_default():
    system = System()
    assert system.role == MessageRole.SYSTEM
    assert system.sender == "system"
    assert system.recipient == "N/A"
    assert system.system_info == DEFAULT_SYSTEM


def test_system_init_custom():
    custom_message = "Custom system message"
    system = System(
        custom_message, sender=SysUtil.id(), recipient=SysUtil.id()
    )
    assert system.system_info == custom_message


def test_system_init_with_datetime():
    system = System(system_datetime=True)
    assert "System Date:" in system.system_info


def test_system_init_with_message_load():
    protected_params = {
        "role": MessageRole.SYSTEM,
        "content": Note(system_info="Test"),
        "sender": "system",
    }
    system = System(
        MessageFlag.MESSAGE_LOAD,
        sender=MessageFlag.MESSAGE_LOAD,
        recipient=MessageFlag.MESSAGE_LOAD,
        system_datetime=MessageFlag.MESSAGE_LOAD,
        protected_init_params=protected_params,
    )
    assert system.role == MessageRole.SYSTEM
    assert system.system_info == "Test"


def test_system_init_with_message_clone():
    system = System(
        MessageFlag.MESSAGE_CLONE,
        sender=MessageFlag.MESSAGE_CLONE,
        recipient=MessageFlag.MESSAGE_CLONE,
        system_datetime=MessageFlag.MESSAGE_CLONE,
    )
    assert system.role == MessageRole.SYSTEM


# Edge cases and additional tests
def test_system_empty_message():
    system = System("")
    assert system.system_info == DEFAULT_SYSTEM


def test_system_none_message():
    system = System(None)
    assert system.system_info == DEFAULT_SYSTEM


def test_system_very_long_message():
    long_message = "a" * 10000
    system = System(long_message)
    assert len(system.system_info) == 10000


def test_system_with_special_characters():
    special_chars = "!@#$%^&*()_+{}[]|\\:;\"'<>,.?/~`"
    system = System(special_chars)
    assert system.system_info == special_chars


def test_system_with_unicode():
    unicode_message = "你好世界 - Hello World - こんにちは世界"
    system = System(unicode_message)
    assert system.system_info == unicode_message


def test_system_serialization():
    system = System("Test message", system_datetime=True)
    serialized = system.to_dict()
    deserialized = System.from_dict(serialized)
    assert deserialized.system_info == system.system_info
    assert deserialized.sender == system.sender


def test_system_clone():
    original = System("Original message")
    cloned = original.clone()
    assert cloned.system_info == original.system_info
    assert cloned.ln_id != original.ln_id


# Performance test
def test_system_performance():
    import time

    start_time = time.time()
    for _ in range(1000):
        System("Test message", system_datetime=True)
    end_time = time.time()
    assert (
        end_time - start_time < 1
    )  # Should create 1000 System messages in less than 1 second


# Thread safety test
def test_system_thread_safety():
    import threading

    def create_system():
        System(f"Thread {threading.get_ident()}", system_datetime=True)

    threads = [threading.Thread(target=create_system) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()


# Test with invalid UTF-8 sequences
def test_system_invalid_utf8():
    invalid_utf8 = b"Invalid UTF-8: \xff"
    with pytest.raises(UnicodeDecodeError):
        System(invalid_utf8.decode("utf-8"))


# Test system with different datetime formats
@pytest.mark.parametrize(
    "datetime_format",
    [
        "2023-01-01",
        "2023-01-01 12:00:00",
        "Jan 1, 2023",
        "1/1/2023",
    ],
)
def test_system_various_datetime_formats(datetime_format):
    system = System("Test", system_datetime=datetime_format)
    assert datetime_format in system.system_info


# Test system with very large datetime string
def test_system_large_datetime():
    large_datetime = "a" * 1000000
    system = System("Test", system_datetime=large_datetime)
    assert large_datetime in system.system_info


# Test for potential memory leaks
def test_system_memory_usage():
    import sys

    initial_size = sys.getsizeof(System("Test"))
    large_system = System("a" * 1000000)
    large_size = sys.getsizeof(large_system)
    assert (
        large_size - initial_size < 2000000
    )  # Ensure no significant unexpected memory increase


# Test system with empty strings for all parameters
def test_system_empty_strings():
    system = System("", sender="", recipient="")
    assert system.system_info == DEFAULT_SYSTEM
    assert system.sender == "system"
    assert system.recipient == "N/A"


# Test system with None values for sender and recipient
def test_system_none_values():
    system = System("Test", sender=None, recipient=None)
    assert system.sender == "system"
    assert system.recipient == "N/A"


# Test system with various types of content
@pytest.mark.parametrize(
    "content",
    [
        123,
        3.14,
        True,
        ["list", "item"],
        {"key": "value"},
    ],
)
def test_system_various_content_types(content):
    system = System(content)
    assert str(content) in system.system_info



import asyncio
from collections import deque
from typing import Any

from lionabc import BaseManager

from lion_core.communication.mail import Mail, Package
from lion_core.generic.exchange import Exchange
from lion_core.generic.pile import Pile, pile
from lion_core.generic.utils import to_list_type
from lion_core.sys_utils import SysUtil


class MailManager(BaseManager):
    """Manages mail operations for multiple sources in the Lion framework."""

    def __init__(self, sources: list[Any] = None) -> None:
        """Initialize a MailManager instance."""
        self.sources: Pile = pile()
        self.mails: dict[str, dict[str, deque]] = {}
        self.execute_stop: bool = False

        if sources:
            self.add_sources(sources)

    def add_sources(self, sources: Any, /) -> None:
        """Add new sources to the MailManager."""
        try:
            sources = to_list_type(sources)
            self.sources.include(sources)
            for item in sources:
                self.mails[item.ln_id] = {}
        except Exception as e:
            raise ValueError("Failed to add source.") from e

    @staticmethod
    def create_mail(
        sender: str,
        recipient: str,
        category: str,
        package: Any,
        request_source=None,
    ) -> Mail:
        """Create a new Mail object."""
        pack = Package(
            category=category, package=package, request_source=request_source
        )
        return Mail(sender=sender, recipient=recipient, package=pack)

    def delete_source(self, source_id: str) -> None:
        """Delete a source from the MailManager."""
        if source_id not in self.sources:
            raise ValueError(f"Source {source_id} does not exist.")
        self.sources.pop(source_id)
        self.mails.pop(source_id)

    def collect(self, sender: str) -> None:
        """Collect mail from a specific sender."""
        if sender not in self.sources:
            raise ValueError(f"Sender source {sender} does not exist.")
        mailbox: Exchange = (
            self.sources[sender]
            if isinstance(self.sources[sender], Exchange)
            else self.sources[sender].mailbox
        )
        while mailbox.pending_outs.size() > 0:
            mail_id = mailbox.pending_outs.popleft()
            mail: Mail = mailbox.pile.pop(mail_id)
            if mail.recipient not in self.sources:
                rec_ = mail.recipient
                raise ValueError(f"Recipient source {rec_} does not exist")
            if mail.sender not in self.mails[mail.recipient]:
                self.mails[mail.recipient].update({mail.sender: deque()})
            self.mails[mail.recipient][mail.sender].append(mail)

    def send(self, recipient: str) -> None:
        """Send mail to a specific recipient."""
        if recipient not in self.sources:
            raise ValueError(f"Recipient source {recipient} does not exist.")
        if not self.mails[recipient]:
            return
        for key in list(self.mails[recipient].keys()):
            pending_mails = self.mails[recipient].pop(key)
            mailbox: Exchange = (
                self.sources[recipient]
                if isinstance(self.sources[recipient], Exchange)
                else self.sources[recipient].mailbox
            )
            while pending_mails:
                mail = pending_mails.popleft()
                mailbox.include(mail, direction="in")

    def collect_all(self) -> None:
        """Collect mail from all sources."""
        for source in self.sources:
            self.collect(sender=SysUtil.get_id(source))

    def send_all(self) -> None:
        """Send mail to all recipients."""
        for source in self.sources:
            self.send(recipient=SysUtil.get_id(source))

    async def execute(self, refresh_time: int = 1) -> None:
        """Execute mail collection and sending process asynchronously."""
        while not self.execute_stop:
            self.collect_all()
            self.send_all()
            await asyncio.sleep(refresh_time)


# File: lion_core/communication/mail_manager.py



from typing import Any

from lionabc import Condition
from pydantic import BaseModel, ConfigDict, Field


class EdgeCondition(BaseModel, Condition):
    """Represents a condition associated with an edge in the Lion framework.

    This class combines Condition characteristics with Pydantic's
    BaseModel for robust data validation and serialization.

    Attributes:
        source (Any): The source for condition evaluation.
    """

    source: Any = Field(
        default=None,
        title="Source",
        description="The source for condition evaluation",
    )

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
    )


__all__ = ["EdgeCondition"]

# File: lion_core/generic/edge_condition.py



from lion_core.converter import ConverterRegistry, JsonConverter


class ComponentConverterRegistry(ConverterRegistry):
    pass


ComponentConverterRegistry.register(JsonConverter)

__all__ = ["ComponentConverterRegistry"]



from lion_core.communication.assistant_response import AssistantResponse


def handle_assistant(
    sender,
    recipient,
    assistant_response,
):
    if isinstance(assistant_response, AssistantResponse):
        return assistant_response

    if assistant_response:
        return AssistantResponse(
            assistant_response=assistant_response,
            sender=sender,
            recipient=recipient,
        )



from typing import Any

from lionabc import Communicatable
from lionabc.exceptions import LionIDError, LionValueError
from pydantic import Field, field_validator

from lion_core.generic.element import Element
from lion_core.sys_utils import SysUtil


def validate_sender_recipient(value: Any, /) -> str:
    """Validate the sender and recipient fields for mail-like communication."""
    if value in ["system", "user", "N/A", "assistant"]:
        return value

    if value is None:
        return "N/A"

    try:
        return SysUtil.get_id(value)
    except LionIDError as e:
        raise LionValueError("Invalid sender or recipient") from e


class BaseMail(Element, Communicatable):
    """Base class for mail-like communication in the LION system.

    Attributes:
        sender: The ID of the sender node.
        recipient: The ID of the recipient node.
    """

    sender: str = Field(
        default="N/A",
        title="Sender",
        description="The ID of the sender node, or 'system', 'user', "
        "or 'assistant'.",
    )

    recipient: str = Field(
        default="N/A",
        title="Recipient",
        description="The ID of the recipient node, or 'system', 'user', "
        "or 'assistant'.",
    )

    @field_validator("sender", "recipient", mode="before")
    @classmethod
    def _validate_sender_recipient(cls, value: Any) -> str:
        """Validate the sender and recipient fields.

        Args:
            value: The value to validate for the sender or recipient.

        Returns:
            The validated sender or recipient ID.

        Raises:
            LionValueError: If the value is not a valid sender or recipient.
        """
        return validate_sender_recipient(value)


# File: lion_core/communication/base.py



from collections.abc import Callable
from typing import Any

from lionabc.exceptions import LionTypeError, LionValueError
from lionfuncs import ucall

from lion_core.action.action_executor import ActionExecutor
from lion_core.action.action_processor import ActionProcessor
from lion_core.form.base import BaseForm
from lion_core.form.form import Form
from lion_core.generic.progression import Progression
from lion_core.rule.base import Rule
from lion_core.rule.rulebook import RuleBook


class RuleProcessor(ActionProcessor):

    event_type: type[Rule] = Rule

    def __init__(
        self,
        capacity: int,
        refresh_time: float,
        rulebook: RuleBook = None,
        strict_rules: bool = True,
        structure_func: Callable = None,
    ):
        super().__init__(capacity, refresh_time)
        self.rulebook = rulebook or RuleBook()
        self.strict_rules = strict_rules
        self.structure_func = structure_func

    def init_rule(
        self,
        rule_order: list = None,
        rule_progress: str = None,  # an existing progress in rulebook
    ):
        if not rule_order:
            rule_order = self.rulebook.default_rule_order

        rule_progress = self.rulebook.rule_flow._find_prog(rule_progress)
        for i in rule_order:
            self.rulebook.init_rule(i, progress=rule_progress)

    async def process_field(
        self,
        field: str,
        value: Any,
        /,
        *,
        field_annotation=None,
        form: BaseForm = None,
        rule_progress=None,
        check_func: Callable = None,
        **kwargs,
    ):
        if field_annotation is None:
            if isinstance(form, BaseForm) and field in form.all_fields:
                field_annotation = form.field_getattr(field, "annotation")

        for i in self.rulebook.rule_flow[rule_progress]:
            rule: Rule = self.rulebook.active_rules[i]
            if await rule.apply(
                field,
                value,
                annotation=field_annotation,
                check_func=check_func,
                **kwargs,
            ):
                try:
                    return await rule.validate(value)
                except Exception as e:
                    raise LionValueError(
                        f"Failed to validate field: {field}"
                    ) from e

        if self.strict_rules:
            error_message = (
                f"Failed to validate {field} because no rule applied."
                " To return the original value directly when no rule "
                "applies, set strict=False."
            )
            raise LionValueError(error_message)

        return value

    async def process(
        self,
        form: Form,
        response: dict | str,
        rule_progress: Progression | str = None,
        structure_str: bool = True,
        structure_func: Callable | None = None,
        **kwargs,  # additional kwargs for fallback_structure
    ):
        return await self.process_form(
            form,
            response,
            rule_progress=rule_progress,
            structure_str=structure_str,
            structure_func=structure_func,
            **kwargs,
        )

    async def process_form(
        self,
        form: Form,
        model_response: dict | str,
        rule_progress=None,
        structure_str: bool = True,
        structure_func: Callable | None = None,
        **kwargs,  # additional kwargs for fallback_structure
    ):
        if isinstance(model_response, str):
            if len(form.request_fields) == 1:
                model_response = {form.request_fields[0]: model_response}
            else:
                if structure_str:
                    structure_func = structure_func or self.structure_func

                    if structure_func is None:
                        raise ValueError(
                            "Response is a string, you asked to structure "
                            "the string but no structure function is provided"
                        )
                    try:
                        model_response = await ucall(
                            structure_func, model_response, **kwargs
                        )
                    except Exception as e:
                        raise ValueError(
                            "Failed to structure the response string"
                            "Response is a string, but form has multiple"
                            " fields to be filled"
                        ) from e

            if not isinstance(model_response, dict):
                raise LionTypeError(
                    expected_type=dict,
                    received_type=type(model_response),
                )

        dict_ = {}
        for k, v in model_response.items():
            if k in form.request_fields:
                kwargs = form.validation_kwargs.get(k, {})
                _annotation = form.field_getattr(k, "annotation")

                kwargs["annotation"] = _annotation
                if (keys := form.field_getattr(k, "keys", None)) is not None:
                    kwargs["keys"] = keys

                v = await self.process_field(
                    k, v, rule_progress=rule_progress, **kwargs
                )

            dict_[k] = v

        form.fill_request_fields(**dict_)
        return form


class RuleExecutor(ActionExecutor):
    pass



import copy
import os
from collections.abc import Sequence
from datetime import datetime, timezone
from hashlib import sha256
from typing import Literal, TypeVar

from lionabc import Observable
from lionabc.exceptions import LionIDError
from lionfuncs import insert_random_hyphens

from lion_core.setting import (
    DEFAULT_LION_ID_CONFIG,
    DEFAULT_TIMEZONE,
    LionIDConfig,
)

T = TypeVar("T")


class SysUtil:
    """Utility class providing various system-related functionalities."""

    @staticmethod
    def time(
        *,
        tz: timezone = DEFAULT_TIMEZONE,
        type_: Literal["timestamp", "datetime", "iso", "custom"] = "timestamp",
        sep: str | None = "T",
        timespec: str | None = "auto",
        custom_format: str | None = None,
        custom_sep: str | None = None,
    ) -> float | str | datetime:
        """
        Get current time in various formats.

        Args:
            tz: Timezone for the time (default: utc).
            type_: Type of time to return (default: "timestamp").
                Options: "timestamp", "datetime", "iso", "custom".
            sep: Separator for ISO format (default: "T").
            timespec: Timespec for ISO format (default: "auto").
            custom_format: Custom strftime format string for
                type_="custom".
            custom_sep: Custom separator for type_="custom",
                replaces "-", ":", ".".

        Returns:
            Current time in the specified format.

        Raises:
            ValueError: If an invalid type_ is provided or if custom_format
                is not provided when type_="custom".
        """
        now = datetime.now(tz=tz)

        match type_:
            case "iso":
                return now.isoformat(sep=sep, timespec=timespec)

            case "timestamp":
                return now.timestamp()

            case "datetime":
                return now

            case "custom":
                if not custom_format:
                    raise ValueError(
                        "custom_format must be provided when type_='custom'"
                    )
                formatted_time = now.strftime(custom_format)
                if custom_sep is not None:
                    for old_sep in ("-", ":", "."):
                        formatted_time = formatted_time.replace(
                            old_sep, custom_sep
                        )
                return formatted_time

            case _:
                raise ValueError(
                    f"Invalid value <{type_}> for `type_`, must be"
                    " one of 'timestamp', 'datetime', 'iso', or 'custom'."
                )

    @staticmethod
    def copy(obj: T, /, *, deep: bool = True, num: int = 1) -> T | list[T]:
        """
        Create one or more copies of an object.

        Args:
            obj: The object to be copied.
            deep: If True, create a deep copy. Otherwise, create a shallow
                copy.
            num: The number of copies to create.

        Returns:
            A single copy if num is 1, otherwise a list of copies.

        Raises:
            ValueError: If num is less than 1.
        """
        if num < 1:
            raise ValueError("Number of copies must be at least 1")

        copy_func = copy.deepcopy if deep else copy.copy
        return (
            [copy_func(obj) for _ in range(num)] if num > 1 else copy_func(obj)
        )

    def id(
        config: LionIDConfig = DEFAULT_LION_ID_CONFIG,
        n: int = None,
        prefix: str = None,
        postfix: str = None,
        random_hyphen: bool = None,
        num_hyphens: int = None,
        hyphen_start_index: int = None,
        hyphen_end_index: int = None,
    ) -> str:
        _dict = {
            "n": n,
            "prefix": prefix,
            "postfix": postfix,
            "random_hyphen": random_hyphen,
            "num_hyphens": num_hyphens,
            "hyphen_start_index": hyphen_start_index,
            "hyphen_end_index": hyphen_end_index,
        }
        _dict = {k: v for k, v in _dict.items() if v is not None}
        config = {**config.to_dict(), **_dict}
        return SysUtil._id(**config)

    @staticmethod
    def _id(
        *,
        n: int,
        prefix: str = "",
        postfix: str = "",
        random_hyphen: bool = False,
        num_hyphens: int = 0,
        hyphen_start_index: int = 6,
        hyphen_end_index: int = -6,
    ) -> str:
        """
        Generate a unique identifier.

        Args:
            n: Length of the ID (excluding prefix and postfix).
            prefix: String to prepend to the ID.
            postfix: String to append to the ID.
            random_hyphen: If True, insert random hyphens into the ID.
            num_hyphens: Number of hyphens to insert if random_hyphen is True.
            hyphen_start_index: Start index for hyphen insertion.
            hyphen_end_index: End index for hyphen insertion.

        Returns:
            A unique identifier string.
        """
        _t = SysUtil.time(type_="iso").encode()
        _r = os.urandom(16)
        _id = sha256(_t + _r).hexdigest()[:n]

        if random_hyphen:
            _id = insert_random_hyphens(
                s=_id,
                num_hyphens=num_hyphens,
                start_index=hyphen_start_index,
                end_index=hyphen_end_index,
            )

        if prefix:
            _id = f"{prefix}{_id}"
        if postfix:
            _id = f"{_id}{postfix}"

        return _id

    @staticmethod
    def get_id(
        item: Sequence[Observable] | Observable | str,
        config: LionIDConfig = DEFAULT_LION_ID_CONFIG,
        /,
    ) -> str:
        """
        Get the Lion ID of an item.

        Args:
            item: The item to get the ID from.
            config: Configuration dictionary for ID validation.

        Returns:
            The Lion ID of the item.

        Raises:
            LionIDError: If the item does not contain a valid Lion ID.
        """

        item_id = None
        if isinstance(item, Sequence) and len(item) == 1:
            item = item[0]

        if isinstance(item, Observable):
            item_id: str = item.ln_id
        else:
            item_id = item

        check = isinstance(item_id, str)
        if check:
            id_len = (
                (len(config.prefix) if config.prefix else 0)
                + config.n
                + config.num_hyphens
                + (len(config.postfix) if config.postfix else 0)
            )
            if len(item_id) != id_len:
                check = False
        if check and config.prefix:
            if item_id.startswith(config.prefix):
                item_id = item_id[len(config.prefix) :]  # noqa
            else:
                check = False
        if check and config.postfix:
            if item_id.endswith(config.postfix):
                item_id = item_id[: -len(config.postfix)]
            else:
                check = False
        if check and config.num_hyphens:
            if config.num_hyphens != item_id.count("-"):
                check = False
        if check and config.hyphen_start_index:
            idx = config.hyphen_start_index - len(config.prefix)
            if idx > 0 and "-" in item_id[:idx]:
                check = False
        if check and config.hyphen_end_index:
            if config.hyphen_end_index < 0:
                idx = config.hyphen_end_index + id_len
            idx -= len(config.prefix + config.postfix)
            if idx < 0 and "-" in item_id[idx:]:
                check = False

        if check:
            return config.prefix + item_id + config.postfix
        if (
            isinstance(item_id, str) and len(item_id) == 32
        ):  # for backward compatibility
            return item_id
        raise LionIDError(
            f"The input object of type <{type(item).__name__}> does "
            "not contain or is not a valid Lion ID. Item must be an instance"
            " of `Observable` or a valid `ln_id`."
        )

    @staticmethod
    def is_id(
        item: Sequence[Observable] | Observable | str,
        config: LionIDConfig = DEFAULT_LION_ID_CONFIG,
        /,
    ) -> bool:
        """
        Check if an item is a valid Lion ID.

        Args:
            item: The item to check.
            config: Configuration dictionary for ID validation.

        Returns:
            True if the item is a valid Lion ID, False otherwise.
        """
        try:
            SysUtil.get_id(item, config)
            return True
        except LionIDError:
            return False



from typing import Any

from lionabc.exceptions import LionValueError
from pydantic import Field, field_validator
from typing_extensions import override

from lion_core.communication.base_mail import BaseMail
from lion_core.communication.package import Package, PackageCategory


class Mail(BaseMail):
    """A mail component with sender, recipient, and package."""

    sender: str = Field(
        ...,
        title="Sender",
        description="The ID of the sender node, or 'system', 'user', "
        "or 'assistant'.",
    )

    recipient: str = Field(
        ...,
        title="Recipient",
        description="The ID of the recipient node, or 'system', 'user', "
        "or 'assistant'.",
    )

    package: Package = Field(
        ...,
        title="Package",
        description="The package to be delivered.",
    )

    @property
    def category(self) -> PackageCategory:
        """Return the category of the package."""
        return self.package.category

    @override
    @field_validator("sender", "recipient", mode="before")
    @classmethod
    def _validate_sender_recipient(cls, value: Any) -> str:
        """Validate the sender and recipient fields."""
        value = super()._validate_sender_recipient(value)
        if value == "N/A":
            raise LionValueError("Invalid sender or recipient for Mail")
        return value


# File: lion_core/communication/mail.py



from datetime import timezone
from enum import Enum
from typing import Any

from lionfuncs import LN_UNDEFINED, LionUndefinedType
from pydantic import BaseModel, ConfigDict, Field


class SchemaModel(BaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True)

    schema_version: int | float | str | None = Field(
        default=None, exclude=True
    )

    def to_dict(self) -> dict[str, Any]:
        dict_ = self.model_dump(exclude_unset=True)
        for i in list(dict_.keys()):
            if dict_[i] is LN_UNDEFINED:
                dict_.pop(i)
        return dict_

    @classmethod
    def from_dict(cls, **data):
        return cls(**data)

    @classmethod
    def schema_keys(cls) -> set:
        return set(cls.model_fields.keys())


class LionIDConfig(SchemaModel):
    n: int
    random_hyphen: bool
    num_hyphens: int
    hyphen_start_index: int
    hyphen_end_index: int
    prefix: str = "ln"
    postfix: str = ""


class TimedFuncCallConfig(SchemaModel):
    initial_delay: int = 0
    retry_default: str | LionUndefinedType = LN_UNDEFINED
    retry_timeout: int | None = None
    retry_timing: bool = False
    error_msg: str | None = None
    error_map: dict | None = None


class RetryConfig(SchemaModel):
    num_retries: int = 0
    initial_delay: int = 0
    retry_delay: int = 0
    backoff_factor: int = 1
    retry_default: str | LionUndefinedType = LN_UNDEFINED
    retry_timeout: int | None = None
    retry_timing: bool = False
    verbose_retry: bool = False
    error_msg: str | None = None
    error_map: dict | None = None


class BaseLionFields(str, Enum):
    LN_ID = "ln_id"
    TIMESTAMP = "timestamp"
    METADATA = "metadata"
    EXTRA_FIELDS = "extra_fields"
    CONTENT = "content"
    CREATED = "created"
    EMBEDDING = "embedding"


class PydanticSerializationConfig(SchemaModel):
    mode: str = "python"
    include: set | None = None
    exclude: set | None = None
    context: dict | None = None
    by_alias: bool = False
    exclude_unset: bool = False
    exclude_defaults: bool = False
    exclude_none: bool = False
    round_trip: bool = False
    warnings: bool = True
    serialize_as_any: bool = False


DEFAULT_LION_ID_CONFIG = LionIDConfig(
    n=42,
    random_hyphen=True,
    num_hyphens=4,
    hyphen_start_index=6,
    hyphen_end_index=-6,
    prefix="ln",
    postfix="",
)

DEFAULT_TIMED_FUNC_CALL_CONFIG = TimedFuncCallConfig(
    initial_delay=0,
    retry_timeout=None,
    retry_timing=False,
    error_msg=None,
    error_map=None,
)

DEFAULT_RETRY_CONFIG = RetryConfig(
    num_retries=0,
    initial_delay=0,
    retry_delay=0,
    backoff_factor=1,
    retry_timeout=None,
    retry_timing=False,
    verbose_retry=False,
    error_msg=None,
    error_map=None,
)


DEFAULT_TIMEZONE = timezone.utc
BASE_LION_FIELDS = set(BaseLionFields.__members__.values())

# File: lion_core/setting.py



from datetime import datetime
from typing import Any, TypeVar

from lionabc.exceptions import LionIDError
from pydantic import ConfigDict, field_validator
from typing_extensions import override

from lion_core._class_registry import LION_CLASS_REGISTRY, get_class
from lion_core.generic.base import RealElement
from lion_core.setting import DEFAULT_TIMEZONE
from lion_core.sys_utils import SysUtil

T = TypeVar("T", bound="Element")


class Element(RealElement):
    """Base class in the Lion framework."""

    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
        use_enum_values=True,
        populate_by_name=True,
    )

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize and register subclasses in the global class registry."""
        super().__pydantic_init_subclass__(**kwargs)
        LION_CLASS_REGISTRY[cls.__name__] = cls

    @property
    def created_datetime(self) -> datetime:
        """Get the creation datetime of the Element."""
        return datetime.fromtimestamp(self.timestamp, tz=DEFAULT_TIMEZONE)

    @field_validator("ln_id", mode="before")
    def _validate_id(cls, value: Any) -> str:
        try:
            return SysUtil.get_id(value)
        except Exception:
            raise LionIDError(f"Invalid lion id: {value}")

    @field_validator("timestamp", mode="before")
    def _validate_timestamp(cls, value: Any) -> float:
        if isinstance(value, str):
            try:
                value = datetime.fromisoformat(value)
            except ValueError:
                raise ValueError(f"Invalid datetime string format: {value}")
        if isinstance(value, datetime):
            return value.timestamp()
        elif isinstance(value, float | int):
            return value
        else:
            raise ValueError(f"Unsupported type for time_attr: {type(value)}")

    @override
    @classmethod
    def from_dict(cls, data: dict, /, **kwargs: Any) -> T:
        """create an instance of the Element or its subclass"""
        if "lion_class" in data:
            cls = get_class(class_name=data.pop("lion_class"))
        if cls.from_dict.__func__ != Element.from_dict.__func__:
            return cls.from_dict(data, **kwargs)
        return cls.model_validate(data, **kwargs)

    @override
    def to_dict(self, **kwargs: Any) -> dict:
        """Convert the Element to a dictionary representation."""
        dict_ = self.model_dump(**kwargs)
        dict_["lion_class"] = self.class_name()
        return dict_

    @override
    def __str__(self) -> str:
        timestamp_str = self.created_datetime.isoformat(timespec="minutes")
        return (
            f"{self.class_name()}(ln_id={self.ln_id[:6]}.., "
            f"timestamp={timestamp_str})"
        )

    def __hash__(self) -> int:
        return hash(self.ln_id)

    def __bool__(self) -> bool:
        """Always True"""
        return True

    def __len__(self) -> int:
        """Return the length of the Element."""
        return 1


__all__ = ["Element"]

# File: lion_core/generic/element.py



import inspect
from enum import Enum
from typing import Any

from lionabc import Relational
from pydantic import Field, field_validator
from typing_extensions import override

from lion_core._class_registry import get_class
from lion_core.communication.base_mail import BaseMail
from lion_core.generic.component import Component
from lion_core.generic.log import Log
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


class MessageField(str, Enum):
    """Enum to store message fields for consistent referencing."""

    LION_ID = "lion_id"
    TIMESTAMP = "timestamp"
    ROLE = "role"
    SENDER = "sender"
    RECIPIENT = "recipient"
    CONTENT = "content"


class MessageRole(str, Enum):
    """Enum for possible roles a message can assume in a conversation."""

    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


class MessageFlag(str, Enum):
    """Enum to signal constructing a clone Message"""

    MESSAGE_CLONE = "MESSAGE_CLONE"
    MESSAGE_LOAD = "MESSAGE_LOAD"


class RoledMessage(Relational, Component, BaseMail):
    """A base class representing a message with roles and properties."""

    content: Note = Field(
        default_factory=Note,
        description="The content of the message.",
    )

    role: MessageRole | None = Field(
        default=None,
        description="The role of the message in the conversation.",
        examples=["system", "user", "assistant"],
    )

    @property
    def image_content(self) -> list[dict[str, Any]] | None:
        """Return image content if present in the message."""
        msg_ = self.chat_msg
        if isinstance(msg_, dict) and isinstance(msg_["content"], list):
            return [i for i in msg_["content"] if i["type"] == "image_url"]
        return None

    @property
    def chat_msg(self) -> dict[str, Any] | None:
        """Return message in chat representation."""
        try:
            return self._format_content()
        except Exception:
            return None

    @field_validator("role")
    def _validate_role(cls, v: Any) -> MessageRole | None:
        """Validates the role of the message."""
        if v in [r.value for r in MessageRole]:
            return MessageRole(v)
        raise ValueError(f"Invalid message role: {v}")

    def _format_content(self) -> dict[str, Any]:
        """Format the message content for chat representation."""
        if self.content.get("images", None):
            content = self.content.to_dict()
        else:
            content = str(self.content.to_dict())
        return {"role": self.role.value, "content": content}

    def clone(self) -> "RoledMessage":
        """Creates a copy of the current RoledMessage object."""
        cls = self.__class__
        signature = inspect.signature(cls.__init__)
        param_num = len(signature.parameters) - 2

        init_args = [MessageFlag.MESSAGE_CLONE] * param_num

        obj = cls(*init_args)
        obj.role = self.role
        obj.content = self.content
        obj.metadata.set("clone_from", self)

        return obj

    @override
    @classmethod
    def from_dict(cls, data: dict, /, **kwargs: Any) -> "RoledMessage":
        """Loads a RoledMessage object from a dictionary."""
        data = SysUtil.copy(data)
        if "lion_class" in data:
            cls = get_class(data.pop("lion_class"))
        signature = inspect.signature(cls.__init__)
        param_num = len(signature.parameters) - 2

        init_args = [MessageFlag.MESSAGE_LOAD] * param_num

        extra_fields = {}
        for k, v in list(data.items()):
            if k not in cls.model_fields:
                extra_fields[k] = data.pop(k)

        obj = cls(*init_args, protected_init_params=data)

        for k, v in extra_fields.items():
            obj.add_field(name=k, value=v)

        metadata = data.get("metadata", {})
        last_updated = metadata.get("last_updated", None)
        if last_updated is not None:
            obj.metadata.set(["last_updated"], last_updated)
        return obj

    @override
    def __str__(self) -> str:
        """Provides a string representation of the message."""
        content_preview = (
            f"{str(self.content)[:75]}..."
            if len(str(self.content)) > 75
            else str(self.content)
        )
        return (
            f"Message(role={self.role}, sender={self.sender}, "
            f"content='{content_preview}')"
        )

    def to_log(self) -> Log:
        dict_ = self.to_dict()
        content = dict_.pop("content")
        _log = Log(
            content=content,
            loginfo=dict_,
        )
        return _log


# File: lion_core/communication/message.py



import time
from datetime import datetime, timezone

import pytest
from lionabc.exceptions import LionValueError
from pydantic_core._pydantic_core import ValidationError

from lion_core.generic.component import Component
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


@pytest.fixture
def valid_id():
    return SysUtil.id()


@pytest.fixture
def timestamp():
    return datetime.now(timezone.utc).timestamp()


def test_basic_initialization(valid_id, timestamp):
    component = Component(ln_id=valid_id, timestamp=timestamp)
    assert component.ln_id == valid_id
    assert component.timestamp == timestamp
    assert isinstance(component.metadata, Note)
    assert component.content is None
    assert component.embedding == []
    assert component.extra_fields == {}


def test_full_initialization(valid_id, timestamp):
    metadata = Note(key="value")
    content = "Test content"
    embedding = [0.1, 0.2, 0.3]
    extra_fields = {"extra_key": "extra_value"}

    component = Component(
        ln_id=valid_id,
        timestamp=timestamp,
        metadata=metadata,
        content=content,
        embedding=embedding,
        extra_fields=extra_fields,
    )

    assert component.metadata.content == {"key": "value"}
    assert component.content == content
    assert component.embedding == embedding
    assert component.extra_fields == extra_fields


def test_metadata_serialization():
    metadata = Note(key="value")
    component = Component(metadata=metadata)
    serialized = component.model_dump()
    assert serialized["metadata"] == {"key": "value"}


def test_add_field():
    component = Component()
    component.add_field("new_field", "new_value")
    assert component.new_field == "new_value"
    assert "new_field" in component.extra_fields

    with pytest.raises(LionValueError):
        component.add_field("new_field", "another_value")


def test_update_field():
    component = Component()
    component.add_field("test_field", "initial_value")
    component.update_field("test_field", "updated_value")
    assert component.test_field == "updated_value"

    component.update_field("new_field", "new_value")
    assert component.new_field == "new_value"


def test_to_dict_consistency(valid_id, timestamp):
    component = Component(
        ln_id=valid_id,
        timestamp=timestamp,
        content="Test content",
        embedding=[0.1, 0.2, 0.3],
    )
    component.add_field("extra_field", "extra_value")

    dict_repr = component.to_dict()

    assert dict_repr["ln_id"] == valid_id
    assert dict_repr["timestamp"] == timestamp
    assert dict_repr["content"] == "Test content"
    assert dict_repr["embedding"] == [0.1, 0.2, 0.3]
    assert dict_repr["extra_field"] == "extra_value"
    assert dict_repr["lion_class"] == "Component"


def test_from_dict(valid_id, timestamp):
    input_dict = {
        "ln_id": valid_id,
        "timestamp": timestamp,
        "content": "Test content",
        "embedding": [0.1, 0.2, 0.3],
        "extra_field": "extra_value",
        "lion_class": "Component",
    }

    component = Component.from_dict(input_dict)

    assert component.ln_id == valid_id
    assert component.timestamp == timestamp
    assert component.content == "Test content"
    assert component.embedding == [0.1, 0.2, 0.3]
    assert component.extra_field == "extra_value"


def test_nested_structures():
    nested_content = {"level1": {"level2": [1, 2, {"level3": "deep"}]}}
    nested_metadata = Note(meta_level1={"meta_level2": ["a", "b", "c"]})

    component = Component(content=nested_content, metadata=nested_metadata)
    serialized = component.to_dict()

    assert serialized["content"] == nested_content
    assert serialized["metadata"] == nested_metadata.content

    deserialized = Component.from_dict(serialized)
    assert deserialized.content == nested_content
    assert deserialized.metadata.content == nested_metadata.content


def test_large_component():
    large_content = "x" * 1000000  # 1MB of data
    large_embedding = [0.1] * 10000  # 10,000 float embeddings

    component = Component(content=large_content, embedding=large_embedding)
    serialized = component.to_dict()
    deserialized = Component.from_dict(serialized)

    assert len(deserialized.content) == 1000000
    assert len(deserialized.embedding) == 10000


def test_component_equality(valid_id):
    comp1 = Component(ln_id=valid_id, content="Same content")
    time.sleep(0.01)  # Ensure different timestamps
    comp2 = Component(ln_id=valid_id, content="Same content")
    comp3 = Component(ln_id=SysUtil.id(), content="Different content")

    assert comp1 != comp2  # Different timestamps
    assert comp1 != comp3  # Different ln_id and content


def test_component_immutability(valid_id, timestamp):
    component = Component(ln_id=valid_id, timestamp=timestamp)

    with pytest.raises(ValidationError):
        component.ln_id = SysUtil.id()

    with pytest.raises(ValidationError):
        component.timestamp = datetime.now().timestamp()


def test_component_with_note_content():
    note_content = Note(key="value")
    component = Component(content=note_content)
    serialized = component.to_dict()

    assert serialized["content"] == {"key": "value"}

    deserialized = Component.from_dict(serialized)
    assert isinstance(deserialized.content, dict)
    assert deserialized.content == {"key": "value"}


def test_component_string_representations(valid_id):
    component = Component(ln_id=valid_id, content="Test content")
    str_repr = str(component)
    repr_repr = repr(component)

    assert valid_id[:8] in str_repr
    assert "Test content" in str_repr
    assert valid_id in repr_repr
    assert "Test content" in repr_repr


# File: tests/test_generic/test_component.py



from typing import Any, Literal

from lionabc import MutableRecord
from lionabc.exceptions import LionValueError
from lionfuncs import LN_UNDEFINED
from pydantic import Field, field_validator
from pydantic_core import PydanticUndefined

from lion_core.form.utils import ERR_MAP
from lion_core.generic.component import Component


class BaseForm(Component, MutableRecord):
    """Base form class providing core functionality for form handling.

    This class serves as a foundation for creating custom forms within the
    lion-core library. It includes methods for managing output fields,
    handling results, and applying field annotations.

    The BaseForm class focuses on output fields, which are fields that are
    presented as the result of form processing. Output fields can include all,
    part, or none of the request fields and can be conditionally modified by
    the process if the form is not set to be strict.

    Attributes:
        assignment: The objective of the task, which may define how input
            fields are processed into output fields.
        template_name: The name of the form template.
        output_fields: A list of field names that are outputted and presented
            by the form.
        none_as_valid_value: Indicates whether to treat None as a valid value
            when processing output fields.
        has_processed: Indicates if the task has been processed.

    Example:
        >>> form = BaseForm(
        ...     assignment="input1, input2 -> output",
        ...     output_fields=["output1", "output2"],
        ... )
        >>> result = form.get_results()
        >>> print(result)
    """

    assignment: str | None = Field(
        default=None,
        description="The objective of the task.",
        examples=["input1, input2 -> output"],
    )
    template_name: str = Field(
        default="default_form", description="Name of the form template"
    )
    output_fields: list[str] = Field(
        default_factory=list,
        description="Fields that are outputted and presented by the form.",
    )
    none_as_valid_value: bool = Field(
        default=False,
        description="Indicate whether to treat None as a valid value.",
    )
    has_processed: bool = Field(
        default=False,
        description="Indicates if the task has been processed.",
        exclude=True,
    )

    def check_is_completed(
        self,
        handle_how: Literal["raise", "return_missing"] = "raise",
    ) -> list[str] | None:
        """Check if all required fields are completed.

        Args:
            handle_how: How to handle incomplete fields.

        Returns:
            List of incomplete fields if handle_how is "return_missing",
            None otherwise.

        Raises:
            ValueError: If required fields are incomplete and handle_how
                is "raise".
        """
        non_complete_request = []
        invalid_values = [LN_UNDEFINED, PydanticUndefined]
        if not self.none_as_valid_value:
            invalid_values.append(None)

        for i in self.required_fields:
            if getattr(self, i) in invalid_values:
                non_complete_request.append(i)

        if non_complete_request:
            if handle_how == "raise":
                raise ERR_MAP["assignment", "incomplete_request"](
                    non_complete_request,
                )
            elif handle_how == "return_missing":
                return non_complete_request
        else:
            self.has_processed = True

    def is_completed(self) -> bool:
        """Check if the form is completed.

        Returns:
            True if the form is completed, False otherwise.
        """
        try:
            self.check_is_completed(handle_how="raise")
            return True
        except Exception:
            return False

    @field_validator("output_fields", mode="before")
    @classmethod
    def _validate_output(cls, value):
        """Validate the output_fields attribute.

        Args:
            value: The value to validate.

        Returns:
            The validated value.

        Raises:
            LionValueError: If the value is invalid.
        """
        if isinstance(value, str):
            return [value]
        if isinstance(value, list) and all(
            [i for i in value if isinstance(i, str)],
        ):
            return value
        if not value:
            return []
        raise LionValueError("Invalid output fields.")

    @property
    def work_fields(self) -> list[str]:
        """Get the list of fields that are outputted by the form.

        Returns:
            A list of field names.
        """
        return self.output_fields

    @property
    def work_dict(self) -> dict[str, Any]:
        """Get a dictionary of all work fields and their values.

        Returns:
            A dictionary of field names and their values.
        """
        return {i: getattr(self, i, LN_UNDEFINED) for i in self.work_fields}

    @property
    def required_fields(self) -> list[str]:
        """Get the list of required fields for the form.

        Returns:
            A list of required field names.
        """
        return self.output_fields

    @property
    def required_dict(self) -> dict[str, Any]:
        """Get a dictionary of all required fields and their values.

        Returns:
            A dictionary of required field names and their values.
        """
        dict_ = {}
        for i in self.required_fields:
            dict_[i] = getattr(self, i, LN_UNDEFINED)

        return dict_

    def get_results(
        self,
        suppress: bool = False,
        valid_only: bool = False,
    ) -> dict[str, Any]:
        """Retrieve the results of the form.

        Args:
            suppress: If True, suppress errors for missing fields.
            valid_only: If True, return only valid (non-empty) results.

        Returns:
            A dictionary of field names and their values.

        Raises:
            ValueError: If a required field is missing and suppress is False.
        """
        result = {}
        out_fields = self.output_fields or getattr(self, "request_fields", [])

        for i in out_fields:
            if i not in self.all_fields:
                if not suppress:
                    raise ERR_MAP["field", "missing"](i)
            else:
                result[i] = getattr(self, i, LN_UNDEFINED)

        # add a validator for output fields

        if valid_only:
            invalid_values = [
                LN_UNDEFINED,
                PydanticUndefined,
            ]
            if not self.none_as_valid_value:
                invalid_values.append(None)

            res_ = {}
            for k, v in result.items():
                if v not in invalid_values:
                    res_[k] = v
            result = res_

        return result

    @property
    def display_dict(self) -> dict[str, Any]:
        """
        Get a dictionary of the required fields and their values for display.

        Returns:
            A dictionary of required field names and their values.
        """
        return self.required_dict


__all__ = ["BaseForm"]



from collections.abc import ItemsView, Iterator, ValuesView
from typing import Any

from lionabc import Communicatable, Container
from lionfuncs import LN_UNDEFINED, flatten, nget, ninsert, npop, nset, to_list
from pydantic import BaseModel, ConfigDict, Field, field_serializer
from typing_extensions import override

from lion_core.sys_utils import SysUtil

INDICE_TYPE = str | list[str | int]


class Note(BaseModel, Container):
    """A container for managing nested dictionary data structures."""

    content: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        use_enum_values=True,
        populate_by_name=True,
    )

    def __init__(self, **kwargs: Any) -> None:
        """Initialize a Note instance with the given keyword arguments."""
        super().__init__()
        self.content = kwargs

    @field_serializer("content")
    def _serialize_content(self, value: Any) -> dict[str, Any]:
        """Serialize the content"""

        output_dict = SysUtil.copy(value, deep=True)
        origin_obj = output_dict.pop("clone_from", None)

        if origin_obj and isinstance(origin_obj, Communicatable):
            info_dict = {
                "clone_from_info": {
                    "original_ln_id": origin_obj.ln_id,
                    "original_timestamp": origin_obj.timestamp,
                    "original_sender": origin_obj.sender,
                    "original_recipient": origin_obj.recipient,
                }
            }
            output_dict.update(info_dict)
        return output_dict

    def pop(
        self,
        indices: INDICE_TYPE,
        /,
        default: Any = LN_UNDEFINED,
    ) -> Any:
        """Remove and return an item from the nested structure."""
        indices = to_list(indices, flatten=True, dropna=True)
        return npop(self.content, indices, default)

    def insert(self, indices: INDICE_TYPE, value: Any, /) -> None:
        """Insert a value into the nested structure at the specified indice"""
        indices = to_list(indices, flatten=True, dropna=True)
        ninsert(self.content, indices, value)

    def set(self, indices: INDICE_TYPE, value: Any, /) -> None:
        """Set a value in the nested structure at the specified indice"""
        indices = to_list(indices, flatten=True, dropna=True)

        if self.get(indices, None) is None:
            self.insert(indices, value)
        else:
            nset(self.content, indices, value)

    def get(
        self,
        indices: INDICE_TYPE,
        /,
        default: Any = LN_UNDEFINED,
    ) -> Any:
        """Get a value from the nested structure at the specified indice"""
        indices = to_list(indices, flatten=True, dropna=True)
        return nget(self.content, indices, default)

    def keys(self, /, flat: bool = False, **kwargs: Any) -> list:
        """
        Get the keys of the Note.

        Args:
            flat: If True, return flattened keys.
            kwargs: Additional keyword arguments for flattening
        """
        if flat:
            return flatten(self.content, **kwargs).keys()
        return list(self.content.keys())

    def values(self, /, flat: bool = False, **kwargs: Any) -> ValuesView:
        """
        Get the values of the Note.

        Args:
            flat: If True, return flattened values.
            kwargs: Additional keyword arguments for flattening
        """
        if flat:
            return flatten(self.content, **kwargs).values()
        return self.content.values()

    def items(self, /, flat: bool = False, **kwargs: Any) -> ItemsView:
        """
        Get the items of the Note.

        Args:
            flat: If True, return flattened items.
            kwargs: Additional keyword arguments for flattening
        """
        if flat:
            return flatten(self.content, **kwargs).items()
        return self.content.items()

    def to_dict(self, **kwargs: Any) -> dict[str, Any]:
        """
        Convert the Note to a dictionary.

        kwargs: Additional keyword arguments for BaseModel.model_dump

        Returns:
            A dictionary representation of the Note.
        """
        output_dict = self.model_dump(**kwargs)
        return output_dict["content"]

    def clear(self) -> None:
        """Clear the content of the Note."""
        self.content.clear()

    def update(
        self,
        indices: INDICE_TYPE,
        value: Any,
    ) -> None:
        existing = None
        if not indices:
            existing = self.content
        else:
            existing = self.get(indices, None)

        if existing is None:
            if not isinstance(value, (list, dict)):
                value = [value]
            self.set(indices, value)

        if isinstance(existing, list):
            if isinstance(value, list):
                existing.extend(value)
            else:
                existing.append(value)

        elif isinstance(existing, dict):
            if isinstance(value, self.__class__):
                value = value.content

            if isinstance(value, dict):
                existing.update(value)
            else:
                raise ValueError(
                    "Cannot update a dictionary with a non-dictionary value."
                )

    @classmethod
    def from_dict(cls, kwargs: Any) -> "Note":
        """Create a Note from a dictionary."""
        return cls(**kwargs)

    def __contains__(self, indices: INDICE_TYPE) -> bool:
        """Check if the Note contains the specified indices."""
        return self.content.get(indices, LN_UNDEFINED) is not LN_UNDEFINED

    def __len__(self) -> int:
        """Return the length of the Note's content."""
        return len(self.content)

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the Note's content."""
        return iter(self.content)

    def __next__(self) -> str:
        """Return the next item in the Note's content."""
        return next(iter(self.content))

    @override
    def __str__(self) -> str:
        """Return a string representation of the Note's content."""
        return str(self.content)

    @override
    def __repr__(self) -> str:
        """Return a detailed string representation of the Note's content."""
        return repr(self.content)

    def __getitem__(self, indices: INDICE_TYPE) -> Any:
        """Get an item from the Note using index notation."""
        indices = to_list(indices, flatten=True, dropna=True)
        return self.get(indices)

    def __setitem__(self, indices: INDICE_TYPE, value: Any) -> None:
        """Set an item in the Note using index notation."""
        self.set(indices, value)


def note(**kwargs: Any) -> Note:
    """Create a Note object from keyword arguments."""
    return Note(**kwargs)


__all__ = ["Note", "note"]

# File : lion_core/generic/note.py



import json
from typing import Any, Protocol, TypeVar, runtime_checkable

from lionfuncs import to_dict

from lion_core.generic.element import Element

T = TypeVar("T", bound=Element)


@runtime_checkable
class Converter(Protocol):
    """Protocol for converter objects.

    Subclasses need to implement the following methods:
    - convert_obj_to_sub_dict
    - convert_sub_to_obj_dict
    - to_obj
    """

    _object: str

    @classmethod
    def object(cls) -> str:
        """Return the object type of the converter."""
        return cls._object

    @classmethod
    def from_obj(
        cls,
        subject_class: type[Element],
        object_: Any,
        **kwargs: Any,
    ) -> dict:
        """Convert an object to a dictionary representation.

        Args:
            subject_class: The target Element subclass.
            object_: The object to convert.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary representation of the object.
        """
        return cls.convert_obj_to_sub_dict(object_=object_, **kwargs)

    @classmethod
    def convert_obj_to_sub_dict(cls, object_: Any, **kwargs: Any) -> dict:
        """Process an object into a valid lion obj dictionary.

        Args:
            object_: The object to convert.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary representation of the object.

        Raises:
            NotImplementedError: If not implemented in subclass.
        """
        raise NotImplementedError

    @classmethod
    def convert_sub_to_obj_dict(cls, subject: T, **kwargs) -> dict:
        """Process a dictionary into an object class dictionary.

        Args:
            subject: The subject to convert.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary representation of the subject.

        Raises:
            NotImplementedError: If not implemented in subclass.
        """
        raise NotImplementedError

    @classmethod
    def to_obj(
        cls,
        subject: T,
        *,
        convert_kwargs: dict = {},
        **kwargs: Any,
    ) -> dict:
        """Convert a subject to an object representation.

        Args:
            subject: The subject to convert.
            convert_kwargs: Kwargs for convert_sub_to_obj_dict.
            **kwargs: Additional keyword arguments.

        Returns:
            An object representation of the subject.

        Raises:
            NotImplementedError: If not implemented in subclass.
        """
        raise NotImplementedError


class JsonConverter(Converter):
    """JSON converter implementation."""

    _object = "json"

    @classmethod
    def convert_obj_to_sub_dict(cls, object_: str, **kwargs: Any) -> dict:
        """Convert a JSON string to a dictionary.

        Args:
            object_: JSON string to convert.
            **kwargs: Additional arguments for json.loads.

        Returns:
            A dictionary representation of the JSON string.
        """
        kwargs["str_type"] = "json"
        return to_dict(object_, **kwargs)

    @classmethod
    def convert_sub_to_obj_dict(cls, subject: T, **kwargs: Any) -> dict:
        """Convert a subject to a dictionary.

        Args:
            subject: The subject to convert.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary representation of the subject.
        """
        return subject.to_dict(**kwargs)

    @classmethod
    def to_obj(
        cls,
        subject: T,
        *,
        convert_kwargs: dict = {},
        **kwargs: Any,
    ) -> Any:
        """Convert a subject to a JSON string.

        Args:
            subject: The subject to convert.
            convert_kwargs: Kwargs for convert_sub_to_obj_dict.
            **kwargs: Additional arguments for json.dumps.

        Returns:
            A JSON string representation of the subject.
        """
        _dict = cls.convert_sub_to_obj_dict(subject=subject, **convert_kwargs)
        return json.dumps(obj=_dict, **kwargs)


class ConverterRegistry:
    """Registry for managing converters in the Lion framework."""

    _converters: dict[str, Converter] = {}

    @classmethod
    def registry_object_keys(cls) -> list[str]:
        """Get a list of registered object keys.

        Returns:
            A list of registered object keys.
        """
        return list(cls._converters.keys())

    @classmethod
    def register(cls, converter: type[Converter]) -> None:
        """Register a converter.

        Args:
            converter: The converter class to register.

        Raises:
            ValueError: If the converter is not a subclass of Converter.
        """
        if not isinstance(converter, type(Converter)):
            err_msg = (
                "In order to register the converter, it needs to be a "
                "subclass of the `Converter` protocol. "
            )
            if isinstance(converter, Converter):
                err_msg += (
                    f"The converter value received <{converter.object()}> "
                    "is an instance. Did you mean to register the class?"
                )
            raise ValueError(err_msg)

        cls._converters[converter.object()] = converter()

    @classmethod
    def get(cls, object_key: str, /) -> Converter:
        """Get a converter by its object key.

        Args:
            object_key: The object key of the converter.

        Returns:
            The converter class.

        Raises:
            KeyError: If no converter is found for the given key.
        """
        try:
            return cls._converters[object_key]
        except KeyError:
            raise KeyError(
                f"No converter found for {object_key}. "
                "Check if it is registered."
            )

    @classmethod
    def convert_from(
        cls,
        subject_class: type[Element],
        object_: Any,
        object_key: str = None,
        **kwargs,
    ) -> dict[str, Any]:
        """Convert an object to a dictionary using a specific converter.

        Args:
            subject_class: The target Element subclass.
            obj: The object to convert.
            key: The object key of the converter to use.
            **kwargs: Additional keyword arguments.

        Returns:
            A dictionary representation of the object.
        """
        converter = cls.get(object_key)
        return converter.from_obj(
            subject_class=subject_class,
            object_=object_,
            **kwargs,
        )

    @classmethod
    def convert_to(
        cls,
        subject: Any,
        object_key: str | type,
        **kwargs,
    ) -> Any:
        """Convert a subject to an object using a specific converter.

        Args:
            subject: The subject to convert.
            key: The object key or type of the converter to use.
            **kwargs: Additional keyword arguments.

        Returns:
            The converted object.
        """
        if not isinstance(object_key, str):
            object_key = object_key.__name__
        converter = cls.get(object_key)
        return converter.to_obj(subject=subject, **kwargs)


# File: lionagi/core/converter.py



import pytest
from pydantic import ValidationError

from lion_core.communication.message import MessageRole, RoledMessage
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


def test_roled_message_init():
    msg = RoledMessage(
        role=MessageRole.USER,
        content=Note(text="Hello"),
        sender=SysUtil.id(),
    )
    assert msg.role == MessageRole.USER
    assert msg.content.get(["text"]) == "Hello"


def test_roled_message_init_invalid_role():
    with pytest.raises(ValueError):
        RoledMessage(role="invalid_role", content=Note(text="Hello"))


def test_roled_message_image_content_no_images():
    msg = RoledMessage(role=MessageRole.USER, content=Note(text="No images"))
    assert msg.image_content is None


def test_roled_message_chat_msg():
    msg = RoledMessage(
        role=MessageRole.ASSISTANT, content=Note(text="Response")
    )
    chat_msg = msg.chat_msg
    assert chat_msg["role"] == "assistant"
    assert chat_msg["content"] == "{'text': 'Response'}"


def test_roled_message_chat_msg_with_images():
    content = Note(
        text="Image description",
        images=["base64_image_data"],
        image_detail="low",
    )
    msg = RoledMessage(role=MessageRole.USER, content=content)
    chat_msg = msg.chat_msg
    assert isinstance(chat_msg["content"], dict)
    assert "images" in chat_msg["content"]


def test_roled_message_validate_role():
    assert (
        RoledMessage._validate_role(MessageRole.SYSTEM) == MessageRole.SYSTEM
    )
    with pytest.raises(ValueError):
        RoledMessage._validate_role("invalid_role")


def test_roled_message_format_content():
    msg = RoledMessage(role=MessageRole.USER, content=Note(text="Hello"))
    formatted = msg._format_content()
    assert formatted["role"] == "user"
    assert formatted["content"] == "{'text': 'Hello'}"


def test_roled_message_clone():
    original = RoledMessage(
        role=MessageRole.ASSISTANT,
        content=Note(text="Original"),
        sender=SysUtil.id(),
    )
    cloned = original.clone()
    assert cloned.role == original.role
    assert cloned.content.to_dict() == original.content.to_dict()
    assert cloned.ln_id != original.ln_id
    assert cloned.metadata.get(["clone_from"]) == original


def test_roled_message_from_dict():
    data = {
        "role": "user",
        "content": {"text": "Hello"},
        "recipient": SysUtil.id(),
    }
    msg = RoledMessage(**data)
    assert msg.role == MessageRole.USER
    assert msg.content.get("text") == "Hello"
    assert msg.sender == "N/A"


def test_roled_message_from_dict_invalid_data():
    invalid_data = {"role": "invalid_role", "content": "Not a Note object"}
    with pytest.raises(ValidationError):
        RoledMessage.from_dict(invalid_data)


# def test_roled_message_str_representation():
#     msg = RoledMessage(
#         role=MessageRole.SYSTEM,
#         content=Note(text="System message"),
#         sender="system",
#     )
#     str_repr = str(msg)
#     assert "Message(role=system" in str_repr
#     assert "sender=system" in str_repr
#     assert "content='{'text': 'System message'}'" in str_repr


def test_roled_message_long_content_str():
    long_text = "A" * 100
    msg = RoledMessage(role=MessageRole.USER, content=Note(text=long_text))
    str_repr = str(msg)
    assert len(str_repr) < 200  # Ensure it's truncated
    assert "..." in str_repr


def test_roled_message_empty_content():
    msg = RoledMessage(role=MessageRole.USER, content=Note())
    assert msg.content.to_dict() == {}


def test_roled_message_clone_with_complex_content():
    complex_content = Note(
        text_="Complex",
        list_=[1, 2, 3],
        dict_={"a": 1, "b": 2},
        nested_=Note(inner="Nested"),
    )
    original = RoledMessage(role=MessageRole.USER, content=complex_content)
    cloned = original.clone()
    assert cloned.content.to_dict() == original.content.to_dict()


# Edge case: Handling of very large content
def test_roled_message_large_content():
    large_content = Note(text="A" * 1000000)  # 1 million characters
    msg = RoledMessage(role=MessageRole.ASSISTANT, content=large_content)
    assert (
        len(str(msg)) < 2000
    )  # Ensure string representation is reasonably sized


# Edge case: Ensure proper handling of None values
def test_roled_message_none_values():
    msg = RoledMessage(role=MessageRole.SYSTEM, content=Note(field=None))
    assert msg.content.get("field") is None
    assert "None" in str(msg)


# Test for potential memory leaks in clone method
def test_roled_message_clone_memory():
    import sys

    original = RoledMessage(
        role=MessageRole.USER, content=Note(text="Memory test")
    )
    initial_size = sys.getsizeof(original)
    clones = [original.clone() for _ in range(1000)]
    assert (
        sys.getsizeof(clones[-1]) - initial_size < 1000
    )  # Ensure no significant memory increase


# Test thread safety of clone method
def test_roled_message_clone_thread_safety():
    import threading

    original = RoledMessage(
        role=MessageRole.ASSISTANT, content=Note(text="Thread test")
    )
    clones = []

    def clone_message():
        clones.append(original.clone())

    threads = [threading.Thread(target=clone_message) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert len(clones) == 100
    assert all(c.content.get("text") == "Thread test" for c in clones)


# Test handling of circular references
def test_roled_message_circular_reference():
    content = Note(text="Circular")
    content.set("self_ref", content)
    msg = RoledMessage(role=MessageRole.USER, content=content)
    cloned = msg.clone()
    assert cloned.content.get("self_ref") is cloned.content


# Test for potential issues with Unicode characters
def test_roled_message_unicode_handling():
    unicode_content = Note(text="Unicode test: 你好世界! 🌍")
    msg = RoledMessage(role=MessageRole.USER, content=unicode_content)
    cloned = msg.clone()
    assert cloned.content.get("text") == "Unicode test: 你好世界! 🌍"
    str_repr = str(cloned)
    assert "你好世界" in str_repr and "🌍" in str_repr


# Test handling of very long role names (edge case)
def test_roled_message_long_role_name():
    long_role = "a" * 1000
    with pytest.raises(ValueError):
        RoledMessage(role=long_role, content=Note(text="Long role test"))


# Test for potential issues with time-sensitive operations
def test_roled_message_time_sensitivity():
    import time

    msg = RoledMessage(role=MessageRole.SYSTEM, content=Note(text="Time test"))
    time.sleep(0.1)
    cloned = msg.clone()
    assert cloned.timestamp > msg.timestamp


# Test handling of invalid UTF-8 sequences
def test_roled_message_invalid_utf8():
    invalid_utf8 = b"Invalid UTF-8: \xff"
    with pytest.raises(UnicodeDecodeError):
        RoledMessage(
            role=MessageRole.USER,
            content=Note(text=invalid_utf8.decode("utf-8")),
        )


# Test for potential issues with multi-threaded access
def test_roled_message_multithreaded_access():
    import threading

    msg = RoledMessage(role=MessageRole.ASSISTANT, content=Note(counter=0))
    lock = threading.Lock()

    def increment_counter():
        with lock:
            current = msg.content.get("counter")
            msg.content.set("counter", current + 1)

    threads = [threading.Thread(target=increment_counter) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert msg.content.get("counter") == 100



from collections import deque
from collections.abc import Generator, Mapping
from typing import Any, TypeVar

from lionabc import Collective
from lionabc.exceptions import LionIDError

from lion_core.generic.element import Element
from lion_core.sys_utils import SysUtil

T = TypeVar("T")


def to_list_type(value: Any, /) -> list[Any]:
    """Convert input to a list format"""
    if value is None:
        return []
    if isinstance(value, str):
        return [value] if SysUtil.is_id(value) else []
    if isinstance(value, Element):
        return [value]
    if isinstance(value, Collective):
        return value.values()
    if isinstance(value, Mapping):
        return list(value.values())
    if isinstance(value, list | tuple | set | deque | Generator):
        return list(value)
    return [value]


def validate_order(value: Any, /) -> list[str]:
    """Validate and standardize order representation"""

    try:
        result = []
        for item in to_list_type(value):
            if isinstance(item, str) and SysUtil.is_id(item):
                result.append(item)
            elif isinstance(item, Element):
                result.append(item.ln_id)
            else:
                id_ = SysUtil.get_id(item)
                if id_:
                    result.append(id_)
        return result
    except Exception as e:
        raise LionIDError("Must only contain valid Lion IDs.") from e


# File: lion_core/container/util.py



import inspect
from typing import Any, Literal

from lionabc.exceptions import LionValueError
from lionfuncs import LN_UNDEFINED
from pydantic import Field, model_validator
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from typing_extensions import override

from lion_core.form.base import BaseForm
from lion_core.form.utils import ERR_MAP, get_input_output_fields
from lion_core.generic.component import T
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


class Form(BaseForm):
    """
    Base task form class extending BaseForm with task-specific functionality.

    This class builds on BaseForm to introduce concepts specific to task-based
    forms, including input fields, request fields, and task-related attributes
    It is designed for tasks that require filling in fields through an
    intelligent process, providing the necessary framework for managing and
    validating these fields.

    Key Concepts:
    - input_fields: Fields required to obtain the request fields.
    - request_fields: Fields that need to be filled by an intelligent process.
    - output_fields: Fields for presentation, which may include all, some, or
      none of the request fields. These can be conditionally modified if the
      form is not strict.

    The Form class provides methods for field management, validation, and
    task instruction generation. It supports both strict and flexible modes
    of operation, allowing for immutable fields and assignments when needed.

    Attributes:
        strict (bool): If True, form fields and assignment are immutable.
        guidance (str | dict[str, Any] | None): High-level task guidance that
            can be optimized by AI or provided as instructions.
        input_fields (list[str]): Fields needed to obtain the requested fields.
        request_fields (list[str]): Fields to be filled by an intelligent
            process.
        task (Any): The work to be done, including custom instructions.
        task_description (str | None): A detailed description of the task.
        init_input_kwargs (dict[str, Any]): Initial keyword arguments for
            input fields.

    Example:
        >>> form = Form(
        ...     assignment="input1, input2 -> output",
        ...     strict=True,
        ...     guidance="Complete the task with the given inputs."
        ... )
        >>> form.fill_input_fields(input1="value1", input2="value2")
        >>> print(form.get_results())
    """

    strict: bool = Field(
        default=False,
        description="If True, form fields and assignment are immutable.",
        frozen=True,
    )
    guidance: str | dict[str, Any] | None = Field(
        default=None,
        description="High-level task guidance, optimizable by AI.",
    )
    input_fields: list[str] = Field(
        default_factory=list,
        description="Fields required to obtain the requested fields.",
    )
    request_fields: list[str] = Field(
        default_factory=list,
        description="Fields to be filled by an intelligent process.",
    )
    task: Any = Field(
        default_factory=str,
        description="Work to be done, including custom instructions.",
    )
    task_description: str | None = Field(
        default_factory=str,
        description="Detailed description of the task",
    )
    init_input_kwargs: dict[str, Any] = Field(
        default_factory=dict,
        exclude=True,
    )

    def check_is_completed(
        self,
        handle_how: Literal["return_missing", "raise"] = "raise",
    ) -> list[str] | None:
        """Check if all required fields are completed.

        Args:
            handle_how: How to handle incomplete fields.

        Returns:
            List of incomplete fields if handle_how is "return_missing",
            None otherwise.

        Raises:
            ValueError: If required fields are incomplete and handle_how
                is "raise".
        """
        if self.strict and self.has_processed:
            return
        return super().check_is_completed(handle_how)

    @model_validator(mode="before")
    @classmethod
    def check_input_output_list_omitted(cls, data: Any) -> dict[str, Any]:
        """Validate and process the input data before model creation.

        Args:
            data: The input data for model creation.

        Returns:
            The validated and processed input data.

        Raises:
            ValueError: If the input data is invalid or missing required fields
        """
        if isinstance(data, Note):
            data = data.to_dict()

        if not isinstance(data, dict):
            raise ERR_MAP["type", "not_dict"](data)

        if not data.get("assignment", None):
            raise ERR_MAP["assignment", "no_assignment"]

        if "input_fields" in data:
            raise ERR_MAP["assignment", "explcit_input"]

        if "request_fields" in data:
            raise ERR_MAP["assignment", "explcit_request"]

        if "task" in data:
            raise ERR_MAP["assignment", "explicit_task"]

        fields = get_input_output_fields(data.get("assignment"))
        input_fields, request_fields = fields

        if not input_fields or input_fields == [""]:
            raise ERR_MAP["assignment", "missing_input"]

        elif not request_fields or request_fields == [""]:
            raise ERR_MAP["assignment", "missing_request"]

        data["input_fields"] = input_fields
        data["request_fields"] = request_fields
        data["output_fields"] = data.get("output_fields", request_fields)
        data["init_input_kwargs"] = {}
        data["strict"] = data.get("strict", False)

        for in_ in data["input_fields"]:
            data["init_input_kwargs"][in_] = (
                data.pop(in_, LN_UNDEFINED)
                if in_ not in cls.model_fields
                else data.get(in_, LN_UNDEFINED)
            )

        return data

    @model_validator(mode="after")
    def check_input_output_fields(self) -> "Form":
        """
        Validate and process input and output fields after model creation.

        Returns:
            The validated Form instance.
        """
        for i in self.input_fields:
            if i in self.model_fields:
                self.init_input_kwargs[i] = getattr(self, i)
            else:
                self.add_field(
                    i,
                    value=self.init_input_kwargs.get(i, LN_UNDEFINED),
                )

        for i in self.request_fields:
            if i not in self.all_fields:
                self.add_field(i)

        return self

    @override
    @property
    def work_fields(self) -> list[str]:
        """Return a list of all fields involved in the task."""
        return self.input_fields + self.request_fields

    @override
    @property
    def required_fields(self) -> list[str]:
        """Return a list of all unique required fields."""
        return list(
            set(self.input_fields + self.request_fields + self.output_fields),
        )

    @property
    def validation_kwargs(self):
        """Get validation keyword arguments for each work field."""
        dict_ = {}
        for i in self.work_fields:
            dict_[i] = self.field_getattr(i, "validation_kwargs", {})
        return dict_

    @property
    def instruction_dict(self) -> dict[str, Any]:
        """Return a dictionary with task instruction information."""
        return {
            "context": self.instruction_context,
            "instruction": self.instruction_prompt,
            "request_fields": self.instruction_request_fields,
        }

    @property
    def instruction_context(self) -> str:
        """Generate a description of the form's input fields."""
        a = self.all_fields
        context = "### Input Fields:\n"
        for idx, i in enumerate(self.input_fields):
            context += f"Input No.{idx+1}: {i}\n"
            if getattr(a[i], "description", None):
                context += f"  - description: {a[i].description}.\n"
            context += f"  - value: {getattr(self, i)}.\n"
        return context

    @property
    def instruction_prompt(self) -> str:
        """Generate a task instruction prompt for the form."""
        a = self.all_fields
        prompt = ""
        if "guidance" in a:
            prompt += f"### Overall Guidance:\n{getattr(self, 'guidance')}.\n"

        in_fields = ", ".join(self.input_fields)
        out_fields = ", ".join(self.output_fields)
        prompt += "### Task Instructions:\n"
        prompt += f"1. Provided Input Fields: {in_fields}.\n"
        prompt += f"2. Requested Output Fields: {out_fields}.\n"
        prompt += f"3. Your task:\n{self.task}.\n"

        return prompt

    @property
    def instruction_request_fields(self) -> dict[str, str]:
        """Get descriptions of the form's requested fields."""
        a = self.all_fields

        context = "### Output Fields:\n"
        for idx, i in enumerate(self.request_fields):
            context += f"Input No.{idx+1}: {i}\n"
            if getattr(a[i], "description", None):
                context += f"  - description: {a[i].description}.\n"
            if getattr(a[i], "examples", None):
                context += f"  - examples: {a[i].examples}.\n"

        return context

    @override
    def update_field(
        self,
        field_name: str,
        value: Any = LN_UNDEFINED,
        annotation: Any = LN_UNDEFINED,
        field_obj: FieldInfo | Any = LN_UNDEFINED,
        **kwargs: Any,
    ) -> None:
        """Update a field in the form.

        Extends the base update_field method to also update the
        init_input_kwargs dictionary.

        Args:
            field_name: The name of the field to update.
            value: The value to assign to the field.
            annotation: The type annotation for the field.
            field_obj: The field object containing metadata.
            **kwargs: Additional keyword arguments for field configuration.
        """
        super().update_field(
            field_name,
            value=value,
            annotation=annotation,
            field_obj=field_obj,
            **kwargs,
        )
        self._fill_init_input_kwargs(field_name)

    @override
    def __setattr__(self, field_name: str, value: Any) -> None:
        """Set an attribute of the form.

        This method enforces the strict attribute, preventing modifications
        to certain fields when strict mode is enabled.

        Args:
            field_name: The name of the attribute to set.
            value: The value to assign to the attribute.

        Raises:
            ValueError: If attempting to modify a restricted field
                        in strict mode.
        """
        if self.strict and field_name in {
            "assignment",
            "input_fields",
            "request_fields",
        }:
            raise ERR_MAP["assignment", "strict"](field_name)

        if field_name in {"input_fields", "request_fields"}:
            raise ERR_MAP["field", "modify_input_request_list"]

        if field_name in {"init_input_kwargs"}:
            raise ERR_MAP["field", "modify_restricted"](field_name)

        super().__setattr__(field_name, value)
        self._fill_init_input_kwargs(field_name)

    def _fill_init_input_kwargs(self, field_name: str) -> None:
        """Update init_input_kwargs if the field is an input field."""
        if field_name in self.input_fields:
            self.init_input_kwargs[field_name] = getattr(self, field_name)

    def check_is_workable(
        self,
        handle_how: Literal["raise", "return_missing"] = "raise",
    ) -> list[str] | None:
        """Check if all input fields are filled and the form is workable.

        Args:
            handle_how: How to handle missing inputs.

        Returns:
            List of missing inputs if handle_how is "return_missing",
            None otherwise.

        Raises:
            ValueError: If input fields are missing and handle_how is
                        "raise".
        """
        if self.strict and self.has_processed:
            raise ERR_MAP["assignment", "strict_processed"]

        missing_inputs = []
        invalid_values = [LN_UNDEFINED, PydanticUndefined]
        if not self.none_as_valid_value:
            invalid_values.append(None)

        for i in self.input_fields:
            if getattr(self, i) in invalid_values:
                missing_inputs.append(i)

        if missing_inputs:
            if handle_how == "raise":
                raise ERR_MAP["assignment", "incomplete_input"](missing_inputs)
            elif handle_how == "return_missing":
                return missing_inputs

    def is_completed(self) -> bool:
        """Check if the form has been completed."""
        try:
            self.check_is_completed(handle_how="raise")
            return True
        except Exception:
            return False

    def is_workable(self) -> bool:
        """Check if the form is workable."""
        try:
            self.check_is_workable(handle_how="raise")
            return True
        except Exception:
            return False

    def to_dict(self, *, valid_only: bool = False) -> dict[str, Any]:
        """Convert the form to a dictionary.

        Args:
            valid_only: Whether to include only valid fields in the output.

        Returns:
            A dictionary representation of the form.
        """
        _dict = super().to_dict()
        if not valid_only:
            return _dict

        disallow_values = [LN_UNDEFINED, PydanticUndefined]
        if not self.none_as_valid_value:
            disallow_values.append(None)
        return {k: v for k, v in _dict.items() if v not in disallow_values}

    @override
    @classmethod
    def from_dict(cls, data: dict, **kwargs) -> T:
        """Create a Form instance from a dictionary.

        Args:
            data: The input data for creating the form.
            **kwargs: Additional keyword arguments.

        Returns:
            The created Form instance.
        """
        input_data = SysUtil.copy(data)

        input_data.pop("lion_class", None)
        input_data.pop("input_fields", None)
        input_data.pop("request_fields", None)
        task = input_data.pop("task", "")

        extra_fields = {}
        for k, v in list(input_data.items()):
            if k not in cls.model_fields:
                extra_fields[k] = input_data.pop(k)
        obj = cls.model_validate(input_data, **kwargs)
        obj.task = task
        for k, v in extra_fields.items():
            obj.update_field(field_name=k, value=v)

        metadata = SysUtil.copy(data.get("metadata", {}))
        last_updated = metadata.get("last_updated", None)
        if last_updated is not None:
            obj.metadata.set(["last_updated"], last_updated)
        else:
            obj.metadata.pop(["last_updated"], None)
        return obj

    def fill_input_fields(
        self,
        form: BaseForm | Any = None,
        **value_kwargs,
    ) -> None:
        """Fill the form's input fields with values.

        Args:
            form: A form to copy values from.
            **value_kwargs: Values to use for filling the input fields.

        Raises:
            TypeError: If the provided form is not a BaseForm instance.
        """
        if form is not None and not isinstance(form, BaseForm):
            raise ERR_MAP["type", "not_form_instance"](form)

        for i in self.input_fields:
            if self.none_as_valid_value:
                if getattr(self, i) is not LN_UNDEFINED:
                    continue
                value = value_kwargs.get(i, LN_UNDEFINED)
                if value is LN_UNDEFINED:
                    value = SysUtil.copy(getattr(form, i, LN_UNDEFINED))
                if value is not LN_UNDEFINED:
                    setattr(self, i, value)
            else:
                if getattr(self, i) in [LN_UNDEFINED, None]:
                    value = value_kwargs.get(i)
                    if value in [LN_UNDEFINED, None]:
                        value = SysUtil.copy(getattr(form, i, LN_UNDEFINED))
                    if value not in [LN_UNDEFINED, None]:
                        setattr(self, i, value)

    def fill_request_fields(
        self,
        form: BaseForm = None,
        **value_kwargs,
    ) -> None:
        """Fill the form's request fields with values.

        Args:
            form: A form to copy values from.
            **value_kwargs: Values to use for filling the request fields.

        Raises:
            TypeError: If the provided form is not a BaseForm instance.
        """
        if form is not None and not isinstance(form, BaseForm):
            raise ERR_MAP["type", "not_form_instance"](form)

        for i in self.request_fields:
            if self.none_as_valid_value:
                if getattr(self, i) is not LN_UNDEFINED:
                    continue
                value = value_kwargs.get(i, LN_UNDEFINED)
                if value is LN_UNDEFINED:
                    value = SysUtil.copy(getattr(form, i, LN_UNDEFINED))
                if value is not LN_UNDEFINED:
                    setattr(self, i, value)
            else:
                if getattr(self, i) in [LN_UNDEFINED, None]:
                    value = value_kwargs.get(i)
                    if value in [LN_UNDEFINED, None]:
                        value = SysUtil.copy(getattr(form, i, LN_UNDEFINED))
                    if value not in [LN_UNDEFINED, None]:
                        setattr(self, i, value)

    @classmethod
    def from_form(
        cls,
        form: BaseForm | type[BaseForm],
        guidance: str | dict[str, Any] | None = None,
        assignment: str | None = None,
        strict: bool = False,
        task_description: str | None = None,
        fill_inputs: bool = True,
        none_as_valid_value: bool = False,
        output_fields: list[str] | None = None,
        same_form_output_fields: bool = False,
        **input_value_kwargs,
    ) -> "Form":
        """Create a Form instance from another form.

        Args:
            form: The form to copy from.
            guidance: Guidance for the new form.
            assignment: The assignment for the new form.
            strict: Whether the new form should be strict.
            task_description: A description of the task.
            fill_inputs: Whether to fill input fields.
            none_as_valid_value: Whether to treat None as a valid value.
            output_fields: Output fields for the new form.
            same_form_output_fields: Whether to copy output fields from the
                original form.
            **input_value_kwargs: Values for filling the new form's input
                                fields.

        Returns:
            The created Form instance.

        Raises:
            TypeError: If the provided form is not a BaseForm or its subclass.
            ValueError: If both output_fields and same_form_output_fields are
                        provided.
        """
        if inspect.isclass(form):
            if not issubclass(form, BaseForm):
                raise ERR_MAP["type", "not_form_class"](form)
            form_fields = form.model_fields
        else:
            if not isinstance(form, BaseForm):
                raise ERR_MAP["type", "not_form_instance"](form)
            form_fields = form.all_fields

        if same_form_output_fields:
            if output_fields:
                raise LionValueError(
                    "Cannot provide output_fields and "
                    "same_form_output_fields at the same time."
                )
            output_fields = SysUtil.copy(form.output_fields)

        if not assignment:
            if not getattr(form, "assignment", None):
                raise ERR_MAP["assignment", "no_assignment"]
            assignment = form.assignment

        obj = cls(
            guidance=guidance or getattr(form, "guidance", None),
            assignment=assignment,
            task_description=task_description
            or getattr(
                form,
                "task_description",
                "",
            ),
            none_as_valid_value=none_as_valid_value
            or getattr(form, "none_as_valid_value", False),
            strict=strict or getattr(form, "strict", False),
            output_fields=output_fields,
        )

        for i in obj.work_fields:
            if i not in form_fields:
                raise ERR_MAP["assignment", "invalid_assignment"](i)
            obj.update_field(i, field_obj=form_fields[i])

        if fill_inputs:
            if inspect.isclass(form):
                obj.fill_input_fields(**input_value_kwargs)
            else:
                obj.fill_input_fields(form=form, **input_value_kwargs)

        return obj

    def remove_request_from_output(self) -> None:
        """Remove the request fields from the output fields."""
        for i in self.request_fields:
            if i in self.output_fields:
                self.output_fields.remove(i)

    def _append_to_one(
        self,
        field_name: str,
        field_type: Literal["input", "output", "request"],
        value: Any = LN_UNDEFINED,
        annotation: Any = LN_UNDEFINED,
        field_obj: FieldInfo | Any = LN_UNDEFINED,
        **kwargs,
    ) -> None:
        """Append a field to one of the field lists.

        Args:
            field_name: The name of the field to append.
            field_type: The type of field to append.
            value: The value of the field.
            annotation: The type annotation for the field.
            field_obj: The field object containing metadata.
            **kwargs: Additional keyword arguments for field configuration.

        Raises:
            ValueError: If the field type is invalid or if appending to
                input or request fields in strict mode.
        """

        if self.strict and field_type in {"input", "request"}:
            raise ERR_MAP["assignment", "strict"](field_type)

        config = {
            "field_name": field_name,
            "value": value,
            "annotation": annotation,
            "field_obj": field_obj,
            **kwargs,
        }

        match field_type:
            case "input":
                if field_name not in self.input_fields:
                    self.input_fields.append(field_name)
                if field_name not in self.assignment:
                    self.assignment = f"{field_name}, " + self.assignment

            case "request":
                if field_name not in self.request_fields:
                    self.request_fields.append(field_name)
                    config.pop("value")
                if field_name not in self.assignment:
                    self.assignment += f", {field_name}"

            case "output":
                if field_name not in self.output_fields:
                    self.output_fields.append(field_name)

            case _:
                raise LionValueError(f"Invalid field type {field_type}")

        if (
            any(
                [
                    value is not LN_UNDEFINED,
                    annotation is not LN_UNDEFINED,
                    field_obj is not LN_UNDEFINED,
                    bool(kwargs),
                ]
            )
            or field_name not in self.all_fields
        ):
            self.update_field(**config)

    def append_to_input(
        self,
        field_name: str,
        value: Any = LN_UNDEFINED,
        annotation: Any = LN_UNDEFINED,
        field_obj: FieldInfo | Any = LN_UNDEFINED,
        **kwargs,
    ) -> None:
        """Append a field to the input fields.

        Args:
            field_name: The name of the field to append.
            value: The value of the field.
            annotation: The type annotation for the field.
            field_obj: The field object containing metadata.
            **kwargs: Additional keyword arguments for field configuration.

        Raises:
            ValueError: If appending the field fails.
        """
        try:
            self._append_to_one(
                field_name=field_name,
                field_type="input",
                value=value,
                annotation=annotation,
                field_obj=field_obj,
                **kwargs,
            )
        except Exception as e:
            raise ERR_MAP["field", "error"](
                f"Failed to append {field_name} to input fields."
            ) from e

    def append_to_output(
        self,
        field_name: str,
        value: Any = LN_UNDEFINED,
        annotation: Any = LN_UNDEFINED,
        field_obj: FieldInfo | Any = LN_UNDEFINED,
        **kwargs,
    ) -> None:
        """Append a field to the output fields.

        Args:
            field_name: The name of the field to append.
            value: The value of the field.
            annotation: The type annotation for the field.
            field_obj: The field object containing metadata.
            **kwargs: Additional keyword arguments for field configuration.

        Raises:
            ValueError: If appending the field fails.
        """

        try:
            self._append_to_one(
                field_name=field_name,
                field_type="output",
                value=value,
                annotation=annotation,
                field_obj=field_obj,
                **kwargs,
            )
        except Exception as e:
            raise ERR_MAP["field", "error"](
                f"Failed to append {field_name} to output fields."
            ) from e

    def append_to_request(
        self,
        field_name: str,
        annotation: Any = LN_UNDEFINED,
        field_obj: FieldInfo | Any = LN_UNDEFINED,
        **kwargs,
    ) -> None:
        """Append a field to the request fields.

        Args:
            field_name: The name of the field to append.
            value: The value of the field.
            annotation: The type annotation for the field.
            field_obj: The field object containing metadata.
            **kwargs: Additional keyword arguments for field configuration.

        Raises:
            ValueError: If appending the field fails.
        """
        if "value" in kwargs:
            raise LionValueError("Cannot provide value to request fields.")
        try:
            self._append_to_one(
                field_name=field_name,
                field_type="request",
                annotation=annotation,
                field_obj=field_obj,
                **kwargs,
            )
        except Exception as e:
            raise ERR_MAP["field", "error"](
                f"Failed to append {field_name} to request fields."
            ) from e


__all__ = ["Form"]



from collections.abc import Iterator
from typing import Any

from lionabc import BaseExecutor, EventStatus
from typing_extensions import override

from lion_core.action.action_processor import ActionProcessor
from lion_core.action.base import ObservableAction
from lion_core.generic.pile import Pile, pile
from lion_core.generic.progression import Progression, prog


class ActionExecutor(BaseExecutor):
    """
    Executor class for managing and processing actions.

    This class is responsible for managing a collection of actions, tracking
    their status, and processing them using a specified processor class.

    Attributes:
        processor_config (dict): Configuration for initializing the processor.
        processor_class (Type[ActionProcessor]): Class used to process actions.
        pile (Pile): A collection of actions managed by the executor.
        pending (Progression): A progression tracking the pending actions.
    """

    processor_class: type[ActionProcessor] = ActionProcessor
    strict: bool = True

    @override
    def __init__(self, **kwargs: Any) -> None:
        """
        Initializes the ActionExecutor with the provided configuration.

        Args:
            **kwargs: Configuration parameters for initializing the processor.
        """
        super().__init__(**kwargs)
        self.pile: Pile[ObservableAction] = pile(
            item_type={self.processor_class.event_type},
            strict=self.strict,
        )
        self.pending: Progression = prog()
        self.processor: ActionProcessor = None

    @property
    def pending_action(self) -> Pile:
        """
        Retrieves a pile of all pending actions.

        Returns:
            Pile: A collection of actions that are still pending.
        """
        return pile(
            [i for i in self.pile if i.status == EventStatus.PENDING],
        )

    @property
    def completed_action(self) -> Pile:
        """
        Retrieves a pile of all completed actions.

        Returns:
            Pile: A collection of actions that have been completed.
        """
        return pile(
            [i for i in self.pile if i.status == EventStatus.COMPLETED],
        )

    async def append(self, action: ObservableAction) -> None:
        """
        Appends a new action to the executor.

        Args:
            action (ObservableAction): The action to be added to the pile.
        """
        await self.pile.ainclude(action)
        self.pending.include(item=action)

    @override
    async def forward(self) -> None:
        """Forwards pending actions to the processor."""
        while len(self.pending) > 0:
            action = self.pile[self.pending.popleft()]
            await self.processor.enqueue(action)
        await self.processor.process()

    def __contains__(self, action: ObservableAction | str) -> bool:
        """Checks if an action is present in the pile."""
        return action in self.pile

    def __iter__(self) -> Iterator[ObservableAction]:
        """Returns an iterator over the actions in the pile."""
        return iter(self.pile)


__all__ = ["ActionExecutor"]
# File: lion_core/action/action_executor.py
